#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2019, Kevin Breit (@kbreit) <kevin.breit@kevinbreit.net>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {
    "metadata_version": "1.1",
    "status": ["deprecated"],
    "supported_by": "community",
}

DOCUMENTATION = r"""
author:
  - Kevin Breit (@kbreit)
deprecated:
  alternative: cisco.meraki.networks_webhooks_http_servers
  removed_in: 3.0.0
  why: Updated modules released with increased functionality
description:
  - Configure and query information about webhooks within the Meraki cloud.
extends_documentation_fragment: cisco.meraki.meraki
module: meraki_webhook
notes:
  - Some of the options are likely only used for developers within Meraki.
options:
  name:
    description:
      - Name of webhook.
    type: str
  net_id:
    description:
      - ID of network which configuration is applied to.
    type: str
  net_name:
    aliases:
      - network
    description:
      - Name of network which configuration is applied to.
    type: str
  payload_template_id:
    description:
      - The ID of the payload template. Overrides payload_template_name if passed
        too.
    type: str
  payload_template_name:
    description:
      - The name of the payload template
    type: str
  shared_secret:
    description:
      - Secret password to use when accessing webhook.
    type: str
  state:
    choices:
      - absent
      - present
      - query
    default: query
    description:
      - Specifies whether object should be queried, created/modified, or removed.
    type: str
  test:
    choices:
      - test
    description:
      - Indicates whether to test or query status.
    type: str
  test_id:
    description:
      - ID of webhook test query.
    type: str
  url:
    description:
      - URL to access when calling webhook.
    type: str
  webhook_id:
    description:
      - Unique ID of webhook.
    type: str
short_description: Manage webhooks configured in the Meraki cloud
"""

EXAMPLES = r"""
- name: Create webhook
  meraki_webhook:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    name: Test_Hook
    url: https://webhook.url/
    shared_secret: shhhdonttellanyone
    payload_template_name: Slack (included)
  delegate_to: localhost
- name: Query one webhook
  meraki_webhook:
    auth_key: abc123
    state: query
    org_name: YourOrg
    net_name: YourNet
    name: Test_Hook
  delegate_to: localhost
- name: Query all webhooks
  meraki_webhook:
    auth_key: abc123
    state: query
    org_name: YourOrg
    net_name: YourNet
  delegate_to: localhost
- name: Delete webhook
  meraki_webhook:
    auth_key: abc123
    state: absent
    org_name: YourOrg
    net_name: YourNet
    name: Test_Hook
  delegate_to: localhost
- name: Test webhook
  meraki_webhook:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    test: test
    url: https://webhook.url/abc123
  delegate_to: localhost
- name: Get webhook status
  meraki_webhook:
    auth_key: abc123
    state: present
    org_name: YourOrg
    net_name: YourNet
    test: status
    test_id: abc123531234
  delegate_to: localhost
"""

RETURN = r"""
data:
    description: List of administrators.
    returned: success
    type: complex
    contains:
        id:
          description: Unique ID of webhook.
          returned: success
          type: str
          sample: aHR0cHM6Ly93ZWJob22LnvpdGUvOGViNWI3NmYtYjE2Ny00Y2I4LTlmYzQtND32Mj3F5NzIaMjQ0
        name:
          description: Descriptive name of webhook.
          returned: success
          type: str
          sample: Test_Hook
        networkId:
          description: ID of network containing webhook object.
          returned: success
          type: str
          sample: N_12345
        shared_secret:
          description: Password for webhook.
          returned: success
          type: str
          sample: VALUE_SPECIFIED_IN_NO_LOG_PARAMETER
        url:
          description: URL of webhook endpoint.
          returned: success
          type: str
          sample: https://webhook.url/abc123
        status:
          description: Status of webhook test.
          returned: success, when testing webhook
          type: str
          sample: enqueued
        payloadTemplate:
          description: The payload template used when posting data to the HTTP server.
          returned: success
          type: str
          sample:
            payloadTemplateId: wpt_00001
            name: Meraki (included)
"""

from ansible.module_utils.basic import AnsibleModule, json
from ansible_collections.cisco.meraki.plugins.module_utils.network.meraki.meraki import (
    MerakiModule,
    meraki_argument_spec,
)


def get_webhook_id(name, webhooks):
    for webhook in webhooks:
        if name == webhook["name"]:
            return webhook["id"]
    return None


def get_all_webhooks(meraki, net_id):
    path = meraki.construct_path("get_all", net_id=net_id)
    response = meraki.request(path, method="GET")
    if meraki.status == 200:
        return response


def sanitize_no_log_values(meraki):
    try:
        meraki.result["diff"]["before"][
            "shared_secret"
        ] = "VALUE_SPECIFIED_IN_NO_LOG_PARAMETER"
    except KeyError:
        pass
    try:
        for i in meraki.result['data']:
            i[
                "shared_secret"
            ] = "VALUE_SPECIFIED_IN_NO_LOG_PARAMETER"
    except (KeyError, TypeError):
        pass
    try:
        meraki.result["data"]["shared_secret"] = "VALUE_SPECIFIED_IN_NO_LOG_PARAMETER"
    except (KeyError, TypeError):
        pass


def get_webhook_payload_templates(meraki, net_id):
    path = meraki.construct_path("get_payload_templates", net_id=net_id)
    response = meraki.request(path, "GET")
    if meraki.status != 200:
        meraki.fail_json(msg="Unable to get webhook payload templates")
    return response


def get_webhook_payload_template_id(meraki, templates, name):
    for template in templates:
        if template["name"] == name:
            return template["payloadTemplateId"]
    meraki.fail_json(msg="No payload template found with the name {0}".format(name))


def main():
    # define the available arguments/parameters that a user can pass to
    # the module

    argument_spec = meraki_argument_spec()
    argument_spec.update(
        state=dict(type="str", choices=["absent", "present", "query"], default="query"),
        net_name=dict(type="str", aliases=["network"]),
        net_id=dict(type="str"),
        name=dict(type="str"),
        url=dict(type="str"),
        shared_secret=dict(type="str", no_log=True),
        webhook_id=dict(type="str"),
        test=dict(type="str", choices=["test"]),
        test_id=dict(type="str"),
        payload_template_name=dict(type="str"),
        payload_template_id=dict(type="str"),
    )

    # the AnsibleModule object will be our abstraction working with Ansible
    # this includes instantiation, a couple of common attr would be the
    # args/params passed to the execution, as well as if the module
    # supports check mode
    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )
    meraki = MerakiModule(module, function="webhooks")

    meraki.params["follow_redirects"] = "all"

    query_url = {"webhooks": "/networks/{net_id}/webhooks/httpServers"}
    query_one_url = {"webhooks": "/networks/{net_id}/webhooks/httpServers/{hookid}"}
    create_url = {"webhooks": "/networks/{net_id}/webhooks/httpServers"}
    update_url = {"webhooks": "/networks/{net_id}/webhooks/httpServers/{hookid}"}
    delete_url = {"webhooks": "/networks/{net_id}/webhooks/httpServers/{hookid}"}
    test_url = {"webhooks": "/networks/{net_id}/webhooks/webhookTests"}
    test_status_url = {"webhooks": "/networks/{net_id}/webhooks/webhookTests/{testid}"}
    query_payload_templates = {
        "webhooks": "/networks/{net_id}/webhooks/payloadTemplates"
    }
    meraki.url_catalog["get_all"].update(query_url)
    meraki.url_catalog["get_one"].update(query_one_url)
    meraki.url_catalog["create"] = create_url
    meraki.url_catalog["update"] = update_url
    meraki.url_catalog["delete"] = delete_url
    meraki.url_catalog["test"] = test_url
    meraki.url_catalog["test_status"] = test_status_url
    meraki.url_catalog["get_payload_templates"] = query_payload_templates

    org_id = meraki.params["org_id"]
    if org_id is None:
        org_id = meraki.get_org_id(meraki.params["org_name"])

    net_id = meraki.params["net_id"]
    if net_id is None:
        nets = meraki.get_nets(org_id=org_id)
        net_id = meraki.get_net_id(net_name=meraki.params["net_name"], data=nets)

    templates = get_webhook_payload_templates(meraki, net_id)
    payload_template_id = meraki.params["payload_template_id"]
    if (
        payload_template_id is None
        and meraki.params["payload_template_name"] is not None
    ):
        payload_template_id = get_webhook_payload_template_id(
            meraki, templates, name=meraki.params["payload_template_name"]
        )

    webhook_id = meraki.params["webhook_id"]
    webhooks = None
    if webhook_id is None and meraki.params["name"]:
        webhooks = get_all_webhooks(meraki, net_id)
        webhook_id = get_webhook_id(meraki.params["name"], webhooks)

    if meraki.params["state"] == "present" and meraki.params["test"] is None:
        payload = {
            "name": meraki.params["name"],
            "url": meraki.params["url"],
        }
        if meraki.params["shared_secret"] is not None:
            payload["sharedSecret"] = meraki.params["shared_secret"]
        if payload_template_id is not None:
            payload["payloadTemplate"] = {"payloadTemplateId": payload_template_id}

    if meraki.params["state"] == "query":
        if webhook_id is not None:  # Query a single webhook
            path = meraki.construct_path(
                "get_one", net_id=net_id, custom={"hookid": webhook_id}
            )
            response = meraki.request(path, method="GET")
            if meraki.status == 200:
                meraki.result["data"] = response
                sanitize_no_log_values(meraki)
                meraki.exit_json(**meraki.result)
        elif meraki.params["test_id"] is not None:
            path = meraki.construct_path(
                "test_status",
                net_id=net_id,
                custom={"testid": meraki.params["test_id"]},
            )
            response = meraki.request(path, method="GET")
            if meraki.status == 200:
                meraki.result["data"] = response
                sanitize_no_log_values(meraki)
                meraki.exit_json(**meraki.result)
        else:
            path = meraki.construct_path("get_all", net_id=net_id)
            response = meraki.request(path, method="GET")
            if meraki.status == 200:
                meraki.result["data"] = response
                # meraki.fail_json(msg=meraki.result)
                sanitize_no_log_values(meraki)
                meraki.exit_json(**meraki.result)
    elif meraki.params["state"] == "present":
        if meraki.params["test"] == "test":
            payload = {"url": meraki.params["url"]}
            path = meraki.construct_path("test", net_id=net_id)
            response = meraki.request(path, method="POST", payload=json.dumps(payload))
            if meraki.status == 201:
                meraki.result["data"] = response
                meraki.exit_json(**meraki.result)
        if webhook_id is None:  # New webhook needs to be created
            if meraki.check_mode is True:
                meraki.result["data"] = payload
                meraki.result["data"]["networkId"] = net_id
                meraki.result["changed"] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path("create", net_id=net_id)
            response = meraki.request(path, method="POST", payload=json.dumps(payload))
            if meraki.status == 201:
                meraki.result["data"] = response
                meraki.result["changed"] = True
        else:  # Need to update
            path = meraki.construct_path(
                "get_one", net_id=net_id, custom={"hookid": webhook_id}
            )
            original = meraki.request(path, method="GET")
            if meraki.is_update_required(original, payload):
                if meraki.check_mode is True:
                    meraki.generate_diff(original, payload)
                    sanitize_no_log_values(meraki)
                    original.update(payload)
                    meraki.result["data"] = original
                    meraki.result["changed"] = True
                    meraki.exit_json(**meraki.result)
                path = meraki.construct_path(
                    "update", net_id=net_id, custom={"hookid": webhook_id}
                )
                response = meraki.request(
                    path, method="PUT", payload=json.dumps(payload)
                )
                if meraki.status == 200:
                    # Not all fields are included so it needs to be checked to avoid comparing same thing
                    if meraki.is_update_required(original, response):
                        meraki.generate_diff(original, response)
                    sanitize_no_log_values(meraki)
                    meraki.result["data"] = response
                    meraki.result["changed"] = True
            else:
                meraki.result["data"] = original
    elif meraki.params["state"] == "absent":
        if webhook_id is None:  # Make sure it is downloaded
            if webhooks is None:
                webhooks = get_all_webhooks(meraki, net_id)
            webhook_id = get_webhook_id(meraki.params["name"], webhooks)
            if webhook_id is None:
                meraki.fail_json(
                    msg="There is no webhook with the name {0}".format(
                        meraki.params["name"]
                    )
                )
        if webhook_id:  # Test to see if it exists
            if meraki.module.check_mode is True:
                meraki.result["data"] = None
                meraki.result["changed"] = True
                meraki.exit_json(**meraki.result)
            path = meraki.construct_path(
                "delete", net_id=net_id, custom={"hookid": webhook_id}
            )
            response = meraki.request(path, method="DELETE")
            if meraki.status == 204:
                meraki.result["data"] = response
                meraki.result["changed"] = True

    # in the event of a successful module execution, you will want to
    # simple AnsibleModule.exit_json(), passing the key/value results
    meraki.exit_json(**meraki.result)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Felix Fontein
# Copyright (c) 2020 Markus Bergholz <markuman+spambelongstogoogle@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this module util is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

# https://dns.hetzner.com/api-docs

from __future__ import annotations

import json
import time
import typing as t
from collections.abc import Mapping
from urllib.parse import quote

from ansible.module_utils.basic import env_fallback

from ansible_collections.community.dns.plugins.module_utils._argspec import ArgumentSpec
from ansible_collections.community.dns.plugins.module_utils._json_api_helper import (
    ERROR_CODES,
    UNKNOWN_ERROR,
    JSONAPIHelper,
)
from ansible_collections.community.dns.plugins.module_utils._provider import (
    ProviderInformation,
)
from ansible_collections.community.dns.plugins.module_utils._record import DNSRecord
from ansible_collections.community.dns.plugins.module_utils._record_set import (
    DNSRecordSet,
)
from ansible_collections.community.dns.plugins.module_utils._zone import DNSZone
from ansible_collections.community.dns.plugins.module_utils._zone_record_api import (
    NOT_PROVIDED,
    DNSAPIError,
    ZoneRecordAPI,
    filter_records,
)
from ansible_collections.community.dns.plugins.module_utils._zone_record_set_api import (
    ZoneRecordSetAPI,
    filter_record_sets,
)

if t.TYPE_CHECKING:
    from collections.abc import Collection, Sequence  # pragma: no cover

    from .._argspec import OptionProvider  # pragma: no cover
    from .._http import HTTPHelper, HTTPMethod  # pragma: no cover
    from .._provider import AnsibleType  # pragma: no cover
    from .._record import IDNSRecord  # pragma: no cover
    from .._record_set import IDNSRecordSet  # pragma: no cover
    from .._zone_record_api import NotProvidedType  # pragma: no cover


def _create_zone_from_json(source: dict[str, t.Any]) -> DNSZone[str]:
    info = source.copy()
    zone = DNSZone(zone_id=info.pop("id"), name=info.pop("name"))
    if "legacy_ns" in info:
        info["legacy_ns"] = sorted(info["legacy_ns"])
    zone.info = info
    return zone


@t.overload
def _create_record_from_json(
    source: t.Any, *, record_type: str | None = None, has_id: t.Literal[True] = True
) -> DNSRecord[str]: ...


@t.overload
def _create_record_from_json(
    source: t.Any, *, record_type: str | None = None, has_id: t.Literal[False]
) -> DNSRecord[str | None]: ...


def _create_record_from_json(
    source: t.Any, *, record_type: str | None = None, has_id: bool = True
) -> DNSRecord:
    if not isinstance(source, dict):
        raise DNSAPIError(
            f"Unexpected data {source!r} when expecting DNS record as JSON object"
        )
    source = dict(source)
    result = DNSRecord(
        record_id=source.pop("id") if has_id else None,
        record_type=source.pop("type", record_type),
        target=source.pop("value"),
    )
    result.ttl = source.pop("ttl", None)
    name = source.pop("name", None)
    if name == "@":
        name = None
    result.prefix = name
    source.pop("zone_id", None)
    result.extra.update(source)
    return result


def _record_to_json(record: IDNSRecord[str | None], zone_id: str) -> dict[str, t.Any]:
    result: dict[str, t.Any] = {
        "name": record.prefix or "@",
        "value": record.target,
        "type": record.type,
        "zone_id": zone_id,
    }
    if record.ttl is not None:
        result["ttl"] = record.ttl
    return result


def _pagination_query(
    query: dict[str, str] | Sequence[tuple[str, str]] | None,
    *,
    block_size: int,
    page: int,
) -> dict[str, str] | Sequence[tuple[str, str]]:
    if isinstance(query, dict):
        query_map = query.copy()
        query_map["per_page"] = str(block_size)
        query_map["page"] = str(page)
        return query_map
    query_list = list(query) if query else []
    query_list.append(("per_page", str(block_size)))
    query_list.append(("page", str(page)))
    return query_list


class HetznerAPI(ZoneRecordAPI[str, str], JSONAPIHelper):
    def __init__(
        self, http_helper, token, api="https://dns.hetzner.com/api/", debug=False
    ):
        JSONAPIHelper.__init__(self, http_helper, token, api=api, debug=debug)

    def _create_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Auth-API-Token": self._token,
        }

    def _extract_only_error_message(self, result: t.Any) -> str:
        # These errors are not documented, but are what I experienced the API seems to return:
        if not isinstance(result, dict):
            return f"Unexpected content: {result!r}"
        res = ""
        if isinstance(result.get("error"), dict):
            if "message" in result["error"]:
                msg = result["error"]["message"]
                res = f'{res} with error message "{msg}"'
            if "code" in result["error"]:
                code = result["error"]["code"]
                res = f"{res} (error code {code})"
        if result.get("message"):
            msg = result["message"]
            res = f'{res} with message "{msg}"'
        return res

    def _extract_error_message(self, result: t.Any | None) -> str:
        if result is None:
            return ""
        if isinstance(result, dict):
            res = self._extract_only_error_message(result)
            if res:
                return res
        return f" with data: {result}"

    def _validate(
        self,
        *,
        result: t.Any | None = None,
        info: dict[str, str],
        expected: Collection[int] | None = None,
        method: HTTPMethod = "GET",
    ) -> None:
        super()._validate(result=result, info=info, expected=expected, method=method)
        if isinstance(result, dict):
            error = result.get("error")
            if isinstance(error, dict):
                status = error.get("code")
                if status is None:
                    return
                url = info.get("url")  # not present when using open_url
                if expected is not None and status in expected:
                    return
                error_code = ERROR_CODES.get(status, UNKNOWN_ERROR)
                more = self._extract_error_message(result)
                raise DNSAPIError(
                    f"{method} {url} resulted in API error {status} ({error_code}){more}"
                )

    def _list_pagination(
        self,
        url: str,
        data_key: str,
        *,
        query: dict[str, str] | Sequence[tuple[str, str]] | None = None,
        block_size: int = 100,
        accept_404: bool = False,
    ) -> list[t.Any] | None:
        result: list[t.Any] = []
        page = 1
        while True:
            query_ = _pagination_query(query, block_size=block_size, page=page)
            res, info = self._get(
                url,
                query=query_,
                must_have_content=[200],
                expected=[200, 404] if accept_404 and page == 1 else [200],
                require_json_object=True,
            )
            if accept_404 and page == 1 and info["status"] == 404:
                return None
            if not isinstance(res, Mapping):
                raise DNSAPIError(
                    f"GET {url} with query parameters {query_} did not result in JSON object, but {type(res)}"
                )
            result.extend(res[data_key])
            if not isinstance(res.get("meta"), dict) and page == 1:
                return result
            if page >= res["meta"]["pagination"]["last_page"]:
                return result
            page += 1

    def get_zone_by_name(self, name: str) -> DNSZone[str] | None:
        """
        Given a zone name, return the zone contents if found.

        @param name: The zone name (string)
        @return The zone information (DNSZone), or None if not found
        """
        result, dummy = self._get("v1/zones", expected=[200, 404], query={"name": name})
        if not isinstance(result, Mapping):
            raise DNSAPIError(
                f"Retrieving zone by name resulted in {type(result)} instead of object"
            )
        for zone in result["zones"]:
            if zone.get("name") == name:
                return _create_zone_from_json(zone)
        return None

    def get_zone_by_id(self, zone_id: str) -> DNSZone[str] | None:
        """
        Given a zone ID, return the zone contents if found.

        @param zone_id: The zone ID
        @return The zone information (DNSZone), or None if not found
        """
        result, info = self._get(
            f"v1/zones/{zone_id}",
            expected=[200, 404],
            must_have_content=[200],
        )
        if not isinstance(result, Mapping):
            raise DNSAPIError(
                f"Retrieving zone by name resulted in {type(result)} instead of object"
            )
        if info["status"] == 404:
            return None
        return _create_zone_from_json(result["zone"])

    def get_zone_records(
        self,
        zone_id: str,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> list[DNSRecord[str]] | None:
        """
        Given a zone ID, return a list of records, optionally filtered by the provided criteria.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return A list of DNSrecord objects, or None if zone was not found
        """
        result = self._list_pagination(
            "v1/records",
            data_key="records",
            query={"zone_id": zone_id},
            accept_404=True,
        )
        if result is None:
            return None
        return filter_records(
            [_create_record_from_json(record) for record in result],
            prefix=prefix,
            record_type=record_type,
        )

    def add_record(
        self, zone_id: str, record: IDNSRecord[str | None]
    ) -> DNSRecord[str]:
        """
        Adds a new record to an existing zone.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The created DNS record (DNSRecord)
        """
        data = _record_to_json(record, zone_id=zone_id)
        result, info = self._post(
            "v1/records", data=data, expected=[200, 422], require_json_object=True
        )
        if info["status"] == 422:
            raise DNSAPIError(
                f'The new {record.type} record with value "{record.target}" and TTL {record.ttl}'
                f" has not been accepted by the server{self._extract_only_error_message(result)}"
            )
        return _create_record_from_json(result["record"])

    def update_record(self, zone_id: str, record: DNSRecord[str]) -> DNSRecord[str]:
        """
        Update a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return The DNS record (DNSRecord)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to update record!")
        data = _record_to_json(record, zone_id=zone_id)
        result, info = self._put(
            f"v1/records/{record.id}",
            data=data,
            expected=[200, 422],
            require_json_object=True,
        )
        if info["status"] == 422:
            raise DNSAPIError(
                f'The updated {record.type} record with value "{record.target}" and TTL {record.ttl}'
                f" has not been accepted by the server{self._extract_only_error_message(result)}"
            )
        return _create_record_from_json(result["record"])

    def delete_record(self, zone_id: str, record: DNSRecord[str]) -> bool:
        """
        Delete a record.

        @param zone_id: The zone ID
        @param record: The DNS record (DNSRecord)
        @return True in case of success (boolean)
        """
        if record.id is None:
            raise DNSAPIError("Need record ID to delete record!")
        dummy, info = self._delete(
            f"v1/records/{record.id}",
            must_have_content=False,
            expected=[200, 404],
        )
        return info["status"] == 200

    @staticmethod
    def _append(results_per_zone_id, zone_id: str, result):
        if zone_id not in results_per_zone_id:
            results_per_zone_id[zone_id] = []
        results_per_zone_id[zone_id].append(result)

    def add_records(
        self,
        records_per_zone_id: Mapping[str, Sequence[IDNSRecord[str | None]]],
        stop_early_on_errors: bool = True,
    ) -> dict[
        str,
        list[
            tuple[DNSRecord[str], t.Literal[True], None]
            | tuple[IDNSRecord[str | None], t.Literal[False], DNSAPIError | None]
        ],
    ]:
        """
        Add new records to an existing zone.

        @param records_per_zone_id: Maps a zone ID to a list of DNS records (DNSRecord)
        @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                     This might only work on some APIs.
        @return A dictionary mapping zone IDs to lists of tuples ``(record, created, failed)``.
                Here ``created`` indicates whether the record was created (``True``) or not (``False``).
                If it was created, ``record`` contains the record ID and ``failed`` is ``None``.
                If it was not created, ``failed`` should be a ``DNSAPIError`` instance indicating why
                it was not created. It is possible that the API only creates records if all succeed,
                in that case ``failed`` can be ``None`` even though ``created`` is ``False``.
        """
        json_records = []
        for zone_id, records in records_per_zone_id.items():
            for record in records:
                json_records.append(_record_to_json(record, zone_id=zone_id))
        data = {"records": json_records}
        # Error 422 means that at least one of the records was not valid
        result, dummy = self._post(
            "v1/records/bulk", data=data, expected=[200, 422], require_json_object=True
        )
        results_per_zone_id: dict[
            str,
            list[
                tuple[DNSRecord[str], t.Literal[True], None]
                | tuple[IDNSRecord[str | None], t.Literal[False], DNSAPIError | None]
            ],
        ] = {}
        # This is the list of invalid records that was detected before accepting the whole set
        for json_record in result.get("invalid_records") or []:
            record = _create_record_from_json(json_record, has_id=False)
            zone_id = json_record["zone_id"]
            self._append(
                results_per_zone_id,
                zone_id,
                (
                    record,
                    False,
                    DNSAPIError(
                        f'Creating {record.type} record "{record.target}" with TTL {record.ttl} for zone {zone_id} failed with unknown reason'
                    ),
                ),
            )
        # This is the list of valid records that were not processed
        for json_record in result.get("valid_records") or []:
            record = _create_record_from_json(json_record, has_id=False)
            zone_id = json_record["zone_id"]
            self._append(results_per_zone_id, zone_id, (record, False, None))
        # This is the list of correctly processed records
        for json_record in result.get("records") or []:
            record = _create_record_from_json(json_record)
            zone_id = json_record["zone_id"]
            self._append(results_per_zone_id, zone_id, (record, True, None))
        return results_per_zone_id

    def update_records(
        self,
        records_per_zone_id: Mapping[str, Sequence[DNSRecord[str]]],
        stop_early_on_errors: bool = True,
    ) -> dict[str, list[tuple[DNSRecord[str], bool, DNSAPIError | None]]]:
        """
        Update multiple records.

        @param records_per_zone_id: Maps a zone ID to a list of DNS records (DNSRecord)
        @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                     This might only work on some APIs.
        @return A dictionary mapping zone IDs to lists of tuples ``(record, updated, failed)``.
                Here ``updated`` indicates whether the record was updated (``True``) or not (``False``).
                If it was not updated, ``failed`` should be a ``DNSAPIError`` instance. If it was
                updated, ``failed`` should be ``None``.  It is possible that the API only updates
                records if all succeed, in that case ``failed`` can be ``None`` even though
                ``updated`` is ``False``.
        """
        # Currently Hetzner's bulk update API seems to be broken, it always returns the error message
        # "An invalid response was received from the upstream server". That's why for now, we always
        # fall back to the default implementation.
        if True:  # pylint: disable=using-constant-test
            return super().update_records(
                records_per_zone_id, stop_early_on_errors=stop_early_on_errors
            )

        json_records = []
        for zone_id, records in records_per_zone_id.items():
            for record in records:
                json_records.append(_record_to_json(record, zone_id=zone_id))
        data = {"records": json_records}
        result, dummy = self._put(
            "v1/records/bulk", data=data, expected=[200], require_json_object=True
        )
        results_per_zone_id = {}
        for json_record in result.get("failed_records") or []:
            record = _create_record_from_json(json_record)
            zone_id = json_record["zone_id"]
            self._append(
                results_per_zone_id,
                zone_id,
                (
                    record,
                    False,
                    DNSAPIError(
                        f'Updating {record.type} record #{record.id} "{record.target}" with TTL {record.ttl} for zone {zone_id} failed with unknown reason'
                    ),
                ),
            )
        for json_record in result.get("records") or []:
            record = _create_record_from_json(json_record)
            zone_id = json_record["zone_id"]
            self._append(results_per_zone_id, zone_id, (record, True, None))
        return results_per_zone_id


def _create_zone_from_new_json(source: dict[str, t.Any]) -> DNSZone[str]:
    info = source.copy()
    # Converting ID to str so both APIs can use same interface
    zone = DNSZone(zone_id=str(info.pop("id")), name=info.pop("name"))
    zone.info = info
    return zone


def _create_record_set_from_new_json(
    source: t.Any, *, record_type: str | None = None
) -> DNSRecordSet[str, str]:
    source = dict(source)
    result: DNSRecordSet[str, str] = DNSRecordSet(
        record_set_id=source.pop("id"), record_type=source.pop("type")
    )
    result.ttl = source.pop("ttl", None)
    name = source.pop("name", None)
    if name == "@":
        name = None
    result.prefix = name
    records = source.pop("records")
    for record in records:
        rec = DNSRecord(
            record_id=result.id, record_type=result.type, target=record["value"]
        )
        rec.prefix = result.prefix
        rec.type = result.type
        rec.ttl = result.ttl
        if record["comment"]:
            rec.extra["comment"] = record["comment"]
        result.records.append(rec)
    source.pop("zone", None)
    result.extra.update(source)
    return result


def _get_creation_json_data(
    record_set: IDNSRecordSet[str | None, str | None],
) -> dict[str, t.Any]:
    return {
        "name": record_set.prefix or "@",
        "type": record_set.type,
        "ttl": record_set.ttl,
        "records": [
            {
                "value": record.target,
                "comment": record.extra.get("comment"),
            }
            for record in record_set.records
        ],
    }


def _get_update_json_data_ttl(
    record_set: IDNSRecordSet[str | None, str | None],
) -> dict[str, t.Any]:
    return {
        "ttl": record_set.ttl,
    }


def _get_update_json_data_value(
    record_set: IDNSRecordSet[str | None, str | None],
) -> dict[str, t.Any]:
    return {
        "records": [
            {
                "value": record.target,
                "comment": record.extra.get("comment"),
            }
            for record in record_set.records
        ],
    }


def _q(value: t.Any) -> str:
    return quote(str(value), safe="@*")


def _get_rrset_url(zone_id: str, prefix: str | None, record_type: str) -> str:
    return f"v1/zones/{_q(zone_id)}/rrsets/{_q(prefix or '@')}/{_q(record_type)}"


def _format_action_error(action_with_error) -> str | None:
    error = action_with_error.get("error")
    if not error:
        return None
    return f"{error['message']} ({error['code']})"


# Number of actions to query together at most
_LIMIT_ACTION_QUERYING = 30


class _HetznerNewAPI(ZoneRecordSetAPI[str, str, str], JSONAPIHelper):
    def __init__(
        self, http_helper, token, api="https://api.hetzner.cloud/", debug=False
    ):
        JSONAPIHelper.__init__(self, http_helper, token, api=api, debug=debug)

    def _create_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._token}",
        }

    def _create_post_headers(self) -> dict[str, str]:
        return {
            "Accept": "application/json",
            "Authorization": f"Bearer {self._token}",
            "Content-Type": "application/json",
        }

    def _extract_only_error_message(self, result: dict[str, t.Any]) -> str:
        res = ""
        if isinstance(result.get("error"), dict):
            msg = result["error"]["message"]
            code = result["error"]["code"]
            res = f'{res} with error message "{msg}" (error code {code})'
            if result["error"]["details"]:
                res = f"{res}. Details: {result['error']['details']}"
        return res

    def _extract_error_message(self, result: t.Any | None) -> str:
        if result is None:
            return ""
        if isinstance(result, dict):
            res = self._extract_only_error_message(result)
            if res:
                return res
        return f" with data: {result}"

    def _is_rate_limiting_result(
        self, content: bytes | None, info: dict[str, t.Any]
    ) -> bool | int | float:
        if info["status"] != 429:
            return False
        if content is None:
            return False
        try:
            result = json.loads(content.decode("utf8"))
        except Exception:
            return False
        if not isinstance(result, dict):
            return False
        error = result.get("error")
        if not isinstance(error, dict):
            return False
        status = error.get("code")
        if status != "rate_limit_exceeded":
            return False
        # TODO: is there a hint how much time we should wait?
        # If yes, adjust check_done_delay accordingly!
        return 5

    def _check_error(
        self,
        method: HTTPMethod,
        url: str,
        result: t.Any,
        accepted: Collection[str] | None = None,
    ) -> str | None:
        if not isinstance(result, dict):
            return None
        error = result.get("error")
        if not isinstance(error, dict):
            return None
        status = error.get("code")
        if accepted is not None and status in accepted:
            return status
        more = self._extract_error_message(result)
        raise DNSAPIError(f"{method} {url} resulted in API error {status}{more}")

    @t.overload
    def _list_pagination(
        self,
        url: str,
        data_key: str,
        *,
        query: dict[str, str] | Sequence[tuple[str, str]] | None = None,
        block_size: int = 100,
        accept_404: t.Literal[False] = False,
    ) -> list[t.Any]: ...

    @t.overload
    def _list_pagination(
        self,
        url: str,
        data_key: str,
        *,
        query: dict[str, str] | Sequence[tuple[str, str]] | None = None,
        block_size: int = 100,
        accept_404: t.Literal[True],
    ) -> list[t.Any] | None: ...

    def _list_pagination(
        self,
        url: str,
        data_key: str,
        *,
        query: dict[str, str] | Sequence[tuple[str, str]] | None = None,
        block_size: int = 100,
        accept_404: bool = False,
    ) -> list[t.Any] | None:
        result = []
        page = 1
        while True:
            query_ = _pagination_query(query, block_size=block_size, page=page)
            res, info = self._get(
                url,
                query=query_,
                must_have_content=[200],
                expected=[200, 404] if accept_404 and page == 1 else [200],
                require_json_object=True,
            )
            if accept_404 and page == 1 and info["status"] == 404:
                return None
            self._check_error(
                "GET",
                url,
                res,
                accepted=["not_found"] if accept_404 and page == 1 else [],
            )
            if not isinstance(res, Mapping):
                raise DNSAPIError(
                    f"GET {url} with query parameters {query_} did not result in JSON object, but {type(res)}"
                )
            result.extend(res[data_key])
            if not isinstance(res.get("meta"), dict) and page == 1:
                return result
            if page >= res["meta"]["pagination"]["last_page"]:
                return result
            page += 1

    def get_zone_by_name(self, name: str) -> DNSZone[str] | None:
        """
        Given a zone name, return the zone contents if found.

        @param name: The zone name (string)
        @return The zone information (DNSZone), or None if not found
        """
        url = f"v1/zones/{_q(name)}"
        result, _info = self._get(url, expected=[200, 404], require_json_object=True)
        if self._check_error("GET", url, result, accepted=["not_found"]) == "not_found":
            return None
        return _create_zone_from_new_json(result["zone"])

    def get_zone_by_id(self, zone_id: str) -> DNSZone[str] | None:
        """
        Given a zone ID, return the zone contents if found.

        @param zone_id: The zone ID
        @return The zone information (DNSZone), or None if not found
        """
        return self.get_zone_by_name(str(zone_id))

    def _get_record_set(
        self, zone_id: str, prefix: str | None, record_type: str
    ) -> DNSRecordSet[str, str] | None:
        url = _get_rrset_url(zone_id, prefix, record_type)
        result, _info = self._get(url, expected=[200, 404], require_json_object=True)
        if self._check_error("GET", url, result, accepted=["not_found"]) == "not_found":
            return None
        return _create_record_set_from_new_json(result["rrset"])

    def get_zone_record_sets(
        self,
        zone_id: str,
        prefix: str | None | NotProvidedType = NOT_PROVIDED,
        record_type: str | NotProvidedType = NOT_PROVIDED,
    ) -> list[DNSRecordSet[str, str]] | None:
        """
        Given a zone ID, return a list of record sets, optionally filtered by the provided criteria.

        @param zone_id: The zone ID
        @param prefix: The prefix to filter for, if provided. Since None is a valid value,
                       the special constant NOT_PROVIDED indicates that we are not filtering.
        @param record_type: The record type to filter for, if provided
        @return A list of DNSrecordSet objects, or None if zone was not found
        """
        query: dict[str, str] = {}
        if prefix is not NOT_PROVIDED:
            query["name"] = prefix or "@"  # type: ignore  # This is a string
        if record_type is not NOT_PROVIDED:
            query["type"] = record_type  # type: ignore  # This is a string
        rrsets = self._list_pagination(
            f"v1/zones/{_q(zone_id)}/rrsets",
            "rrsets",
            query=query,
            accept_404=True,
        )
        if rrsets is None:
            return None
        return filter_record_sets(
            [_create_record_set_from_new_json(rrset) for rrset in rrsets],
            prefix=prefix,
            record_type=record_type,
        )

    def _wait_for_actions(
        self,
        actions: list[dict[str, t.Any]],
        what: str,
        *,
        fail_on_error: bool = True,
        stop_on_first_error: bool = False,
    ) -> tuple[list[dict[str, t.Any]], list[dict[str, t.Any]]]:
        errors: list[dict[str, t.Any]] = []
        other_actions: list[dict[str, t.Any]] = []
        # Only do an initial wait if there have been at most _LIMIT_ACTION_QUERYING/2 calls.
        # (This is a heuristic to avoid unnecessary waits when we can already clear the list.)
        do_wait = 2 * len(actions) <= _LIMIT_ACTION_QUERYING
        while True:
            new_errors = [action for action in actions if action["status"] == "error"]
            if new_errors:
                errors.extend(new_errors)
                if stop_on_first_error:
                    break
            other_actions.extend(
                action for action in actions if action["status"] != "running"
            )
            actions = [action for action in actions if action["status"] == "running"]
            if not actions:
                break
            if do_wait:
                time.sleep(1)
            do_wait = True
            action_ids = [str(action["id"]) for action in actions]
            if len(action_ids) == 1:
                url = f"v1/actions/{_q(action_ids[0])}"
                result, dummy = self._get(url, expected=[200], require_json_object=True)
                self._check_error("GET", url, result)
                actions = [result["action"]]
            else:
                remaining_actions = []
                if len(action_ids) > _LIMIT_ACTION_QUERYING:
                    action_ids = action_ids[:_LIMIT_ACTION_QUERYING]
                    remaining_actions = actions[_LIMIT_ACTION_QUERYING:]
                url = "v1/actions"
                result, dummy = self._get(
                    url,
                    query=[("id", action_id) for action_id in sorted(action_ids)],
                    expected=[200],
                    require_json_object=True,
                )
                self._check_error("GET", url, result)
                actions = result["actions"]
                if all(action["status"] != "running" for action in actions):
                    # If all actions queried by this request have stopped, don't wait before querying more
                    do_wait = False
                actions.extend(remaining_actions)
        if errors and fail_on_error:
            error_messages = [_format_action_error(error) for error in errors]
            joined_msgs = ", ".join(msg for msg in error_messages if msg) or "unknown"
            raise DNSAPIError(f"Error while {what}: {joined_msgs}")
        return errors, other_actions + actions

    def add_record_set(
        self,
        zone_id: str,
        record_set: IDNSRecordSet[str | None, str | None],
    ) -> DNSRecordSet[str, str]:
        """
        Adds a new record set to an existing zone.

        @param zone_id: The zone ID
        @param record: The DNS record set (DNSRecordSet)
        @return The created DNS record set (DNSRecordSet)
        """
        data = _get_creation_json_data(record_set)
        url = f"v1/zones/{_q(zone_id)}/rrsets"
        result, dummy = self._post(
            url, data=data, expected=[201], require_json_object=True
        )
        self._check_error("POST", url, result)
        self._wait_for_actions([result["action"]], "adding record set")
        return _create_record_set_from_new_json(result["rrset"])

    def update_record_set(
        self,
        zone_id: str,
        record_set: DNSRecordSet[str, str],
        updated_records: bool = True,
        updated_ttl: bool = True,
    ) -> DNSRecordSet[str, str]:
        """
        Update a record set.

        @param zone_id: The zone ID
        @param record_set: The DNS record set (DNSRecordSet)
        @param updated_records: Hint whether the values were updated.
        @param updated_ttl: Hint whether the values were updated.
        @return The DNS record set (DNSRecordSet)
        """
        actions = []
        base_url = _get_rrset_url(zone_id, record_set.prefix, record_set.type)
        if updated_ttl:
            data = _get_update_json_data_ttl(record_set)
            url = f"{base_url}/actions/change_ttl"
            ttl_result, dummy = self._post(
                url, data=data, expected=[201], require_json_object=True
            )
            self._check_error("POST", url, ttl_result)
            actions.append(ttl_result["action"])
        if updated_records:
            data = _get_update_json_data_value(record_set)
            url = f"{base_url}/actions/set_records"
            set_result, dummy = self._post(
                url, data=data, expected=[201], require_json_object=True
            )
            self._check_error("POST", url, set_result)
            actions.append(set_result["action"])
        self._wait_for_actions(actions, "changing record set")
        result = self._get_record_set(zone_id, record_set.prefix, record_set.type)
        # If the record set suddenly vanished, return the one we *wanted* to have
        return result if result else record_set

    def delete_record_set(
        self, zone_id: str, record_set: DNSRecordSet[str, str]
    ) -> bool:
        """
        Delete a record set.

        @param zone_id: The zone ID
        @param record_set: The DNS record set (DNSRecordSet)
        @return True in case of success (boolean)
        """
        url = _get_rrset_url(zone_id, record_set.prefix, record_set.type)
        result, _info = self._delete(url, expected=[201, 404], require_json_object=True)
        if (
            self._check_error("DELETE", url, result, accepted=["not_found"])
            == "not_found"
        ):
            return False
        self._wait_for_actions([result["action"]], "deleting record set")
        return True

    def _fetch_all_records(
        self, zone_id: str, prefixes_and_types: Sequence[tuple[str | None, str]]
    ) -> list[DNSRecordSet[str, str]]:
        if not prefixes_and_types:
            return []
        common_prefix: str | None | NotProvidedType
        common_type: str | NotProvidedType
        common_prefix, common_type = prefixes_and_types[0]
        if len(prefixes_and_types) == 1:
            res = self._get_record_set(zone_id, common_prefix, common_type)
            return [res] if res else []
        if any(common_prefix != prefix for prefix, _r_type in prefixes_and_types):
            common_prefix = NOT_PROVIDED
        if any(common_type != r_type for _prefix, r_type in prefixes_and_types):
            common_type = NOT_PROVIDED
        return (
            self.get_zone_record_sets(
                zone_id, prefix=common_prefix, record_type=common_type
            )
            or []
        )

    @t.overload
    def _collect_results(
        self,
        what: str,
        data_per_zone_id: dict[
            str,
            list[
                tuple[DNSRecordSet[str, str], list[str], None]
                | tuple[
                    IDNSRecordSet[str | None, str | None],
                    None,
                    DNSAPIError,
                ]
            ],
        ],
        actions: dict[t.Any, dict[str, t.Any]],
        *,
        do_wait: bool,
        stop_early_on_errors: bool,
        refresh_rrsets: bool = True,
        has_incomplete: t.Literal[True],
    ) -> dict[
        str,
        list[
            tuple[DNSRecordSet[str, str], t.Literal[True], None]
            | tuple[
                IDNSRecordSet[str | None, str | None],
                t.Literal[False],
                DNSAPIError | None,
            ]
        ],
    ]: ...

    @t.overload
    def _collect_results(
        self,
        what: str,
        data_per_zone_id: dict[
            str,
            list[tuple[DNSRecordSet[str, str], list[str] | None, DNSAPIError | None]],
        ],
        actions: dict[t.Any, dict[str, t.Any]],
        *,
        do_wait: bool,
        stop_early_on_errors: bool,
        refresh_rrsets: bool = True,
        has_incomplete: t.Literal[False],
    ) -> dict[
        str,
        list[tuple[DNSRecordSet[str, str], bool, DNSAPIError | None]],
    ]: ...

    def _collect_results(
        self,
        what: str,
        data_per_zone_id: Mapping[
            str,
            Sequence[
                tuple[
                    DNSRecordSet[str, str] | IDNSRecordSet[str | None, str | None],
                    list[str] | None,
                    DNSAPIError | None,
                ]
            ],
        ],
        actions: dict[t.Any, dict[str, t.Any]],
        *,
        do_wait: bool,
        stop_early_on_errors: bool,
        refresh_rrsets: bool = True,
        has_incomplete: bool,
    ) -> dict[str, list[t.Any]]:
        errors = {}
        if do_wait:
            action_errors, _other_actions = self._wait_for_actions(
                list(actions.values()),
                what,
                fail_on_error=False,
                stop_on_first_error=stop_early_on_errors,
            )
            for action_error in action_errors:
                errors[action_error["id"]] = DNSAPIError(
                    _format_action_error(action_error)
                )
        results_per_zone_id = {}
        for zone_id, datas in data_per_zone_id.items():
            result: list[
                tuple[
                    DNSRecordSet[str, str] | IDNSRecordSet[str | None, str | None],
                    bool,
                    DNSAPIError | None,
                ]
            ] = []
            results_per_zone_id[zone_id] = result
            refresh = {}
            for record_set, action_ids, error in datas:
                if error is not None:
                    result.append((record_set, False, error))
                error = None
                rr_errors = [
                    errors[action_id]
                    for action_id in action_ids or []
                    if action_id in errors
                ]
                if rr_errors:
                    # If there are multiple ones, only use the first
                    result.append((record_set, False, rr_errors[0]))
                else:
                    # Need to refresh this record set
                    refresh[(record_set.prefix, record_set.type)] = len(result)
                    result.append((record_set, True, None))
            if refresh_rrsets:
                for record_set in self._fetch_all_records(
                    zone_id, list(refresh.keys())
                ):
                    index = refresh.get((record_set.prefix, record_set.type))
                    if index is not None:
                        result[index] = (record_set, True, None)
        return results_per_zone_id

    def add_record_sets(
        self,
        record_sets_per_zone_id: Mapping[
            str, Sequence[IDNSRecordSet[str | None, str | None]]
        ],
        stop_early_on_errors: bool = True,
    ) -> dict[
        str,
        list[
            tuple[DNSRecordSet[str, str], t.Literal[True], None]
            | tuple[
                IDNSRecordSet[str | None, str | None],
                t.Literal[False],
                DNSAPIError | None,
            ]
        ],
    ]:
        """
        Add new record sets to an existing zone.

        @param record_sets_per_zone_id: Maps a zone ID to a list of DNS record sets (DNSRecordSet)
        @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                     This might only work on some APIs.
        @return A dictionary mapping zone IDs to lists of tuples ``(record_set, created, failed)``.
                Here ``created`` indicates whether the record set was created (``True``) or not (``False``).
                If it was created, ``record_set`` contains the record set ID and ``failed`` is ``None``.
                If it was not created, ``failed`` should be a ``DNSAPIError`` instance indicating why
                it was not created. It is possible that the API only creates record sets if all succeed,
                in that case ``failed`` can be ``None`` even though ``created`` is ``False``.
        """
        actions = {}
        data_per_zone_id = {}
        stop = False
        for zone_id, record_sets in record_sets_per_zone_id.items():
            data: list[
                tuple[DNSRecordSet[str, str], list[str], None]
                | tuple[
                    IDNSRecordSet[str | None, str | None],
                    None,
                    DNSAPIError,
                ]
            ] = []
            data_per_zone_id[zone_id] = data
            for record_set in record_sets:
                creation_data = _get_creation_json_data(record_set)
                try:
                    url = f"v1/zones/{_q(zone_id)}/rrsets"
                    res, dummy = self._post(
                        url,
                        data=creation_data,
                        expected=[201],
                        require_json_object=True,
                    )
                    self._check_error("POST", url, res)
                    action = res["action"]
                    actions[action["id"]] = action
                    data.append(
                        (
                            _create_record_set_from_new_json(res["rrset"]),
                            [action["id"]],
                            None,
                        )
                    )
                except DNSAPIError as e:
                    data.append((record_set, None, e))
                    if stop_early_on_errors:
                        stop = True
                        break
            if stop:
                break
        return self._collect_results(
            "adding record sets",
            data_per_zone_id,
            actions,
            do_wait=not stop,
            stop_early_on_errors=stop_early_on_errors,
            refresh_rrsets=False,
            has_incomplete=True,
        )

    def update_record_sets(
        self,
        record_sets_per_zone_id: Mapping[
            str, Sequence[tuple[DNSRecordSet[str, str], bool, bool]]
        ],
        stop_early_on_errors: bool = True,
    ) -> dict[
        str,
        list[tuple[DNSRecordSet[str, str], bool, DNSAPIError | None]],
    ]:
        """
        Update multiple record sets.

        @param record_sets_per_zone_id: Maps a zone ID to a list of tuples
                                        (record_set, updated_records, updated_ttl)
                                        of type (DNSRecordSet, bool, bool).
        @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                     This might only work on some APIs.
        @return A dictionary mapping zone IDs to lists of tuples ``(record_set, updated, failed)``.
                Here ``updated`` indicates whether the record set was updated (``True``) or not (``False``).
                If it was not updated, ``failed`` should be a ``DNSAPIError`` instance. If it was
                updated, ``failed`` should be ``None``.  It is possible that the API only updates
                record sets if all succeed, in that case ``failed`` can be ``None`` even though
                ``updated`` is ``False``.
        """
        actions = {}
        data_per_zone_id = {}
        stop = False
        for zone_id, record_sets in record_sets_per_zone_id.items():
            data: list[
                tuple[DNSRecordSet[str, str], list[str] | None, DNSAPIError | None]
            ] = []
            data_per_zone_id[zone_id] = data
            for record_set, updated_records, updated_ttl in record_sets:
                base_url = _get_rrset_url(zone_id, record_set.prefix, record_set.type)
                try:
                    action_ids = []
                    if updated_ttl:
                        ttl_data = _get_update_json_data_ttl(record_set)
                        ttl_url = f"{base_url}/actions/change_ttl"
                        ttl_result, dummy = self._post(
                            ttl_url,
                            data=ttl_data,
                            expected=[201],
                            require_json_object=True,
                        )
                        self._check_error("POST", ttl_url, ttl_result)
                        action = ttl_result["action"]
                        action_id = action["id"]
                        actions[action_id] = action
                        action_ids.append(action_id)
                    if updated_records:
                        set_data = _get_update_json_data_value(record_set)
                        set_url = f"{base_url}/actions/set_records"
                        set_result, dummy = self._post(
                            set_url,
                            data=set_data,
                            expected=[201],
                            require_json_object=True,
                        )
                        self._check_error("POST", set_url, set_result)
                        action = set_result["action"]
                        action_id = action["id"]
                        actions[action_id] = action
                        action_ids.append(action_id)
                    data.append((record_set, action_ids, None))
                except DNSAPIError as e:
                    data.append((record_set, None, e))
                    if stop_early_on_errors:
                        stop = True
                        break
            if stop:
                break
        return self._collect_results(
            "updating record sets",
            data_per_zone_id,
            actions,
            do_wait=not stop,
            stop_early_on_errors=stop_early_on_errors,
            refresh_rrsets=not stop,
            has_incomplete=False,
        )

    def delete_record_sets(
        self,
        record_sets_per_zone_id: Mapping[str, Sequence[DNSRecordSet[str, str]]],
        stop_early_on_errors: bool = True,
    ) -> dict[
        str,
        list[tuple[DNSRecordSet[str, str], bool, DNSAPIError | None]],
    ]:
        """
        Delete multiple record_sets.

        @param record_sets_per_zone_id: Maps a zone ID to a list of DNS record sets (DNSRecordSet)
        @param stop_early_on_errors: If set to ``True``, try to stop changes after the first error happens.
                                     This might only work on some APIs.
        @return A dictionary mapping zone IDs to lists of tuples ``(record_set, deleted, failed)``.
                In case ``record_set`` was deleted or not deleted, ``deleted`` is ``True``
                respectively ``False``, and ``failed`` is ``None``. In case an error happened
                while deleting, ``deleted`` is ``False`` and ``failed`` is a ``DNSAPIError``
                instance hopefully providing information on the error.
        """
        actions = {}
        data_per_zone_id = {}
        stop = False
        for zone_id, record_sets in record_sets_per_zone_id.items():
            data: list[
                tuple[DNSRecordSet[str, str], list[str] | None, DNSAPIError | None]
            ] = []
            data_per_zone_id[zone_id] = data
            for record_set in record_sets:
                try:
                    url = _get_rrset_url(zone_id, record_set.prefix, record_set.type)
                    result, _info = self._delete(
                        url, expected=[201, 404], require_json_object=True
                    )
                    if (
                        self._check_error("DELETE", url, result, accepted=["not_found"])
                        == "not_found"
                    ):
                        data.append((record_set, None, None))
                    else:
                        action = result["action"]
                        actions[action["id"]] = action
                        data.append((record_set, [action["id"]], None))
                except DNSAPIError as e:
                    data.append((record_set, None, e))
                    if stop_early_on_errors:
                        stop = True
                        break
            if stop:
                break
        return self._collect_results(
            "deleting record sets",
            data_per_zone_id,
            actions,
            do_wait=not stop,
            stop_early_on_errors=stop_early_on_errors,
            refresh_rrsets=False,
            has_incomplete=False,
        )


class HetznerProviderInformation(ProviderInformation):
    def __init__(self, option_provider: OptionProvider | None = None) -> None:
        self._api = None
        if option_provider is not None:
            hetzner_token = option_provider.get_option("hetzner_token")
            hetzner_api_token = option_provider.get_option("hetzner_api_token")
            if hetzner_token is not None:
                self._api = "old"
            if hetzner_api_token is not None:
                self._api = "new"

    def get_supported_record_types(self) -> list[str]:
        """
        Return a list of supported record types.
        """
        old_api = [
            "A",
            "AAAA",
            "NS",
            "MX",
            "CNAME",
            "RP",
            "TXT",
            "SOA",
            "HINFO",
            "SRV",
            "TLSA",
            "DS",
            "CAA",
            "DANE",
        ]
        new_api = [
            "A",
            "AAAA",
            "NS",
            "MX",
            "CNAME",
            "RP",
            "TXT",
            "SOA",
            "HINFO",
            "SRV",
            "TLSA",
            "DS",
            "CAA",
            "HTTPS",
            "PTR",
            "SVCB",
        ]
        if self._api == "old":
            return old_api
        if self._api == "new":
            return new_api
        return sorted(set(old_api) | set(new_api))

    def get_zone_id_type(self) -> AnsibleType:
        """
        Return the (short) type for zone IDs, like ``'int'`` or ``'str'``.
        """
        # For the new API, this would be 'int'.
        # Since we have to support both APIs, we're using 'str'.
        return "str"

    def get_record_id_type(self) -> AnsibleType:
        """
        Return the (short) type for record IDs, like ``'int'`` or ``'str'``.
        """
        return "str"

    def get_record_default_ttl(self) -> int | None:
        """
        Return the default TTL for records, like 300, 3600 or None.
        None means that some other TTL (usually from the zone) will be used.
        """
        return None

    def normalize_prefix(self, prefix: str | None) -> str | None:
        """
        Given a prefix (string or None), return its normalized form.

        The result should always be None for the trivial prefix, and a non-zero length DNS name
        for a non-trivial prefix.

        If a provider supports other identifiers for the trivial prefix, such as '@', this
        function needs to convert them to None as well.
        """
        return None if prefix in ("@", "") else prefix

    def supports_bulk_actions(self) -> bool:
        """
        Return whether the API supports some kind of bulk actions.
        """
        return True

    def txt_record_handling(
        self,
    ) -> t.Literal["decoded", "encoded", "encoded-no-char-encoding"]:
        """
        Return how the API handles TXT records.

        Returns one of the following strings:
        * 'decoded' - the API works with unencoded values
        * 'encoded' - the API works with encoded values
        * 'encoded-no-char-encoding' - the API works with encoded values, but without character encoding
        """
        return "encoded-no-char-encoding"

    def txt_always_quote(self) -> bool:
        """
        Return whether TXT records sent to the API should always be quoted.

        Returns a boolean.

        This return value is only used if txt_record_handling does not return 'decoded'.
        """
        return self._api == "new"


def create_hetzner_provider_information(
    option_provider: OptionProvider | None = None,
) -> HetznerProviderInformation:
    return HetznerProviderInformation(option_provider=option_provider)


def create_hetzner_argument_spec() -> ArgumentSpec:
    return ArgumentSpec(
        argument_spec={
            "hetzner_token": {
                "type": "str",
                "no_log": True,
                "aliases": ["api_token"],
                "fallback": (env_fallback, ["HETZNER_DNS_TOKEN"]),
            },
            "hetzner_api_token": {
                "type": "str",
                "no_log": True,
                "fallback": (env_fallback, ["HETZNER_API_TOKEN"]),
            },
        },
        mutually_exclusive=[["hetzner_token", "hetzner_api_token"]],
        required_one_of=[["hetzner_token", "hetzner_api_token"]],
    )


def create_hetzner_api(
    option_provider: OptionProvider, http_helper: HTTPHelper
) -> HetznerAPI | _HetznerNewAPI:
    hetzner_token = option_provider.get_option("hetzner_token")
    hetzner_api_token = option_provider.get_option("hetzner_api_token")
    if hetzner_token is not None:
        return HetznerAPI(http_helper, hetzner_token)
    if hetzner_api_token is not None:
        return _HetznerNewAPI(http_helper, hetzner_api_token)
    raise AssertionError(
        "One of hetzner_token and hetzner_api_token must be provided"
    )  # pragma: no cover



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2025, Gaspard Micol (@gmicol) <gmicol@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_match_rule_community_term
version_added: "2.12.0"
short_description: Manage Match Community Terms on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage Match Community Terms inside a Match Rule Policy on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v4.1 (NDO v5.1) and later.
author:
- Gaspard Micol (@gmicol)
options:
  template:
    description:
    - The name of the template.
    - The template must be a Tenant Policy template.
    - This parameter or O(template_id) is required.
    type: str
  template_id:
    description:
    - The ID of the template.
    - The template must be a Tenant Policy template.
    - This parameter or O(template) is required.
    type: str
  match_rule_policy:
    description:
    - The name of the Match Rule Policy.
    - This parameter or O(match_rule_policy_uuid) is required.
    type: str
  match_rule_policy_uuid:
    description:
    - The UUID of the Match Rule Policy.
    - This parameter or O(match_rule_policy) is required.
    type: str
  name:
    description:
    - The name of the Match Community Term.
    type: str
    aliases: [ match_community_term ]
  description:
    description:
    - The description of the Match Community Term.
    type: str
    aliases: [ match_community_term_description ]
  match_community_factors:
    description:
    - The list of Match Community Factors.
    type: list
    elements: dict
    suboptions:
      community_factor:
        description:
        - The Community Factor to match.
        - e.g., regular:as2-nn2:4:15, extended:as4-nn2:5:16,
          extended:color:35, no-export, no-advertise, etc.
        type: str
        aliases: [ community ]
      scope:
        description:
        - The scope of the Match Community Factor.
        - Defaults to C(transitive) when unset during creation.
        type: str
        choices: [ transitive, non_transitive ]
      description:
        description:
        - The description of the Match Community Factor.
        type: str
        aliases: [ match_community_description ]
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_template) to create the Tenant Policy template.
- The O(match_rule_policy) must exist before using this module in your playbook.
  Use M(cisco.mso.ndo_match_rule_policy) to create the Match Rule Policy.
seealso:
- module: cisco.mso.ndo_template
- module: cisco.mso.ndo_match_rule_policy
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create a Match Community Term
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    description: Ansible Match Community Term
    match_community_factors:
      - community_factor: no-export
        scope: transitive
        description: no export and transitive
    state: present
  register: create_match_name

- name: Update a Match Community Term using Match Rule Policy's name
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    description: Updated Ansible Match Community Term
    match_community_factors:
      - community_factor: no-export
        scope: non_transitive
        description: no export and non-transitive
      - community_factor: no-advertise
        scope: transitive
        description: no advertise and transitive
    state: present

- name: Update a Match Community Term using Match Rule Policy's UUID
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy_uuid: "{{ create_match_name.current.matchRulePolicyUuid }}"
    name: ansible_match_community_term
    description: Updated Ansible Match Community Term
    match_community_factors:
      - community_factor: no-export
        scope: non_transitive
        description: no export and non-transitive
      - community_factor: no-advertise
        scope: transitive
        description: no advertise and transitive
    state: present

- name: Update a Match Community Term by removing a Match Community
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    description: Updated Ansible Match Community Term
    match_community_factors:
      - community_factor: no-advertise
        scope: transitive
        description: no advertise and transitive
    state: present

- name: Update a Match Community Term by removing all Match Communities
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    description: Updated Ansible Match Community Term
    match_community_factors: []
    state: present

- name: Query a Match Community Term using Match Rule Policy's name
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    state: query
  register: query_with_name

- name: Query a Match Community Term using  Match Rule Policy's UUID
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy_uuid: "{{ create_match_name.current.matchRulePolicyUuid }}"
    name: ansible_match_community_term
    state: query
  register: query_with_uuid

- name: Query all Match Community Terms in a Match Rule Policy
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    state: query
  register: query_all_objects

- name: Delete a Match Community Term using Match Rule Policy's name
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_test
    match_rule_policy: ansible_match_rule_policy
    name: ansible_match_community_term
    state: absent

- name: Delete a Match Community Term using Match Rule Policy's UUID
  cisco.mso.ndo_match_rule_community_term:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template_id: ansible_test
    match_rule_policy_uuid: "{{ create_match_name.current.matchRulePolicyUuid }}"
    name: ansible_match_community_term
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.templates import MSOTemplates
from ansible_collections.cisco.mso.plugins.module_utils.constants import MATCH_COMMUNITY_SCOPE_MAP
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data
import copy


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        template=dict(type="str"),
        template_id=dict(type="str"),
        match_rule_policy=dict(type="str"),
        match_rule_policy_uuid=dict(type="str"),
        name=dict(type="str", aliases=["match_community_term"]),
        description=dict(type="str", aliases=["match_community_term_description"]),
        match_community_factors=dict(
            type="list",
            elements="dict",
            options=dict(
                community_factor=dict(type="str", aliases=["community"]),
                scope=dict(type="str", choices=list(MATCH_COMMUNITY_SCOPE_MAP)),
                description=dict(type="str", aliases=["match_community_description"]),
            ),
        ),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        mutually_exclusive=[
            ("template", "template_id"),
            ["match_rule_policy", "match_rule_policy_uuid"],
        ],
        required_if=[
            ["state", "absent", ["name"]],
            ["state", "present", ["name"]],
        ],
        required_one_of=[
            ["template", "template_id"],
            ["match_rule_policy", "match_rule_policy_uuid"],
        ],
    )

    mso = MSOModule(module)
    mso_templates = MSOTemplates(mso)

    template_name = module.params.get("template")
    template_id = module.params.get("template_id")
    match_rule_policy = module.params.get("match_rule_policy")
    match_rule_policy_uuid = module.params.get("match_rule_policy_uuid")
    name = module.params.get("name")
    description = module.params.get("description")
    match_community_factors = module.params.get("match_community_factors")
    if not isinstance(match_community_factors, list):
        match_community_factors = []
    from_name = module.params.get("from_name")
    to_name = module.params.get("to_name")
    state = module.params.get("state")

    mso_template = mso_templates.get_template("tenant", template_name, template_id)
    mso_template.validate_template("tenantPolicy")

    match_rule_policy_object = mso_template.get_match_rule_policy_object(match_rule_policy_uuid, match_rule_policy)
    if not match_rule_policy_object:
        mso.fail_json(
            msg="The Match Rule Policy with the following {0[0]} '{0[1]}' does not exist".format(
                ("UUID", match_rule_policy_uuid) if match_rule_policy_uuid else ("name", match_rule_policy)
            )
        )
    match_identifiers = {"name": name}
    match = mso_template.get_direct_child_object(match_rule_policy_object, "Match Community Term", "matchCommunityTermsList", match_identifiers)

    if name and match:
        mso.existing = mso.previous = copy.deepcopy(
            mso_template.update_match_rule_policy_child_object_with_template_and_parent(match_rule_policy_object.details, match.details)
        )  # Query a specific object
    elif match:
        mso.existing = [
            mso_template.update_match_rule_policy_child_object_with_template_and_parent(match_rule_policy_object.details, obj) for obj in match
        ]  # Query all objects

    match_name_path = "/tenantPolicyTemplate/template/matchRulePolicies/{0}/matchCommunityTermsList/{1}".format(
        match_rule_policy_object.index,
        match.index if match else "-",
    )

    ops = []

    if state == "present":

        mso_values = {
            "name": name,
            "matchCommunityList": [
                {
                    "community": match_community.get("community_factor"),
                    "scope": MATCH_COMMUNITY_SCOPE_MAP.get(match_community.get("scope")),
                    "description": match_community.get("description"),
                }
                for match_community in match_community_factors
            ],
            "description": description,
            "fromPfxLen": from_name,
            "toPfxLen": to_name,
        }

        if match:
            append_update_ops_data(ops, match.details, match_name_path, mso_values)
            mso.sanitize(mso_values, collate=True)
        else:
            mso.sanitize(mso_values)
            ops.append(dict(op="add", path=match_name_path, value=mso.sent))

    elif state == "absent":
        if match:
            ops.append(dict(op="remove", path=match_name_path))

    if not module.check_mode and ops:
        response = mso.request(mso_template.template_path, method="PATCH", data=ops)
        match_rule_policy_object = mso_template.get_match_rule_policy_object(match_rule_policy_uuid, match_rule_policy, search_object=response)
        match = mso_template.get_direct_child_object(match_rule_policy_object, "Match Community Term", "matchCommunityTermsList", match_identifiers)
        if match:
            mso.existing = mso_template.update_match_rule_policy_child_object_with_template_and_parent(
                match_rule_policy_object.details, match.details
            )  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        if mso.proposed:
            mso_template.update_match_rule_policy_child_object_with_template_and_parent(match_rule_policy_object.details, mso.proposed)
        mso.existing = mso.proposed if state == "present" else {}
    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2018 Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: acme_challenge_cert_helper
author: "Felix Fontein (@felixfontein)"
short_description: Prepare certificates required for ACME challenges such as C(tls-alpn-01)
description:
  - Prepares certificates for ACME challenges such as C(tls-alpn-01).
  - The raw data is provided by the M(community.crypto.acme_certificate) module, and needs to be converted to a certificate
    to be used for challenge validation. This module provides a simple way to generate the required certificates.
seealso:
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the C(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
extends_documentation_fragment:
  - community.crypto._attributes
  - community.crypto._cryptography_dep.minimum
attributes:
  check_mode:
    support: none
    details:
      - This action does not modify state.
  diff_mode:
    support: N/A
    details:
      - This action does not modify state.
  idempotent:
    support: none
    details:
      - The certificates returned are never the same, since the Not Before and Not After timestamps
        depend on the invocation's timestamp.
options:
  challenge:
    description:
      - The challenge type.
    type: str
    required: true
    choices:
      - tls-alpn-01
  challenge_data:
    description:
      - The RV(community.crypto.acme_certificate#module:challenge_data) entry provided by M(community.crypto.acme_certificate)
        for the challenge.
    type: dict
    required: true
  private_key_src:
    description:
      - Path to a file containing the private key file to use for this challenge certificate.
      - Mutually exclusive with O(private_key_content).
    type: path
  private_key_content:
    description:
      - Content of the private key to use for this challenge certificate.
      - Mutually exclusive with O(private_key_src).
    type: str
  private_key_passphrase:
    description:
      - Passphrase to use to decode the private key.
    type: str
    version_added: 1.6.0
"""

EXAMPLES = r"""
---
- name: Create challenges for a given CRT for sample.com
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    challenge: tls-alpn-01
    csr: /etc/pki/cert/csr/sample.com.csr
    dest: /etc/httpd/ssl/sample.com.crt
    modify_account: false
  register: sample_com_challenge

- name: Create certificates for challenges
  community.crypto.acme_challenge_cert_helper:
    challenge: tls-alpn-01
    challenge_data: "{{ item.value['tls-alpn-01'] }}"
    private_key_src: /etc/pki/cert/key/sample.com.key
  loop: "{{ sample_com_challenge.challenge_data | dictsort }}"
  register: sample_com_challenge_certs

- name: Install challenge certificates
  # We need to set up HTTPS such that for the domain,
  # regular_certificate is delivered for regular connections,
  # except if ALPN selects the "acme-tls/1"; then, the
  # challenge_certificate must be delivered.
  # This can for example be achieved with very new versions
  # of NGINX; search for ssl_preread and
  # ssl_preread_alpn_protocols for information on how to
  # route by ALPN protocol.
  ...:
    domain: "{{ item.domain }}"
    challenge_certificate: "{{ item.challenge_certificate }}"
    regular_certificate: "{{ item.regular_certificate }}"
    private_key: /etc/pki/cert/key/sample.com.key
  loop: "{{ sample_com_challenge_certs.results }}"

- name: Create certificate for a given CSR for sample.com
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    challenge: tls-alpn-01
    csr: /etc/pki/cert/csr/sample.com.csr
    dest: /etc/httpd/ssl/sample.com.crt
    data: "{{ sample_com_challenge }}"
    modify_account: false
"""

RETURN = r"""
domain:
  description:
    - The domain the challenge is for. The certificate should be provided if this is specified in the request's the C(Host)
      header.
  returned: always
  type: str
identifier_type:
  description:
    - The identifier type for the actual resource identifier.
  returned: always
  type: str
  choices:
    - dns
    - ip
identifier:
  description:
    - The identifier for the actual resource. Will be a domain name if RV(identifier_type=dns), or an IP address if RV(identifier_type=ip).
  returned: always
  type: str
challenge_certificate:
  description:
    - The challenge certificate in PEM format.
  returned: always
  type: str
regular_certificate:
  description:
    - A self-signed certificate for the challenge domain.
    - If no existing certificate exists, can be used to set-up https in the first place if that is needed for providing the
      challenge.
  returned: always
  type: str
"""

import base64
import datetime
import ipaddress
import typing as t

from ansible.module_utils.basic import AnsibleModule
from ansible.module_utils.common.text.converters import to_bytes, to_text

from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.io import read_file
from ansible_collections.community.crypto.plugins.module_utils._crypto.cryptography_support import (
    CRYPTOGRAPHY_TIMEZONE,
    set_not_valid_after,
    set_not_valid_before,
)
from ansible_collections.community.crypto.plugins.module_utils._cryptography_dep import (
    COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION,
    assert_required_cryptography_version,
)
from ansible_collections.community.crypto.plugins.module_utils._time import (
    get_now_datetime,
)

try:
    import cryptography
    import cryptography.hazmat.backends
    import cryptography.hazmat.primitives.asymmetric.dsa
    import cryptography.hazmat.primitives.asymmetric.ec
    import cryptography.hazmat.primitives.asymmetric.ed448
    import cryptography.hazmat.primitives.asymmetric.ed25519
    import cryptography.hazmat.primitives.asymmetric.padding
    import cryptography.hazmat.primitives.asymmetric.rsa
    import cryptography.hazmat.primitives.asymmetric.utils
    import cryptography.hazmat.primitives.hashes
    import cryptography.hazmat.primitives.serialization
    import cryptography.x509
    import cryptography.x509.oid
except ImportError:
    pass


# Convert byte string to ASN1 encoded octet string
def encode_octet_string(octet_string: bytes) -> bytes:
    if len(octet_string) >= 128:
        raise ModuleFailException(
            "Cannot handle octet strings with more than 128 bytes"
        )
    return bytes([0x4, len(octet_string)]) + octet_string


def main() -> t.NoReturn:
    module = AnsibleModule(
        argument_spec={
            "challenge": {"type": "str", "required": True, "choices": ["tls-alpn-01"]},
            "challenge_data": {"type": "dict", "required": True},
            "private_key_src": {"type": "path"},
            "private_key_content": {"type": "str", "no_log": True},
            "private_key_passphrase": {"type": "str", "no_log": True},
        },
        required_one_of=(["private_key_src", "private_key_content"],),
        mutually_exclusive=(["private_key_src", "private_key_content"],),
    )

    assert_required_cryptography_version(
        module, minimum_cryptography_version=COLLECTION_MINIMUM_CRYPTOGRAPHY_VERSION
    )

    try:
        # Get parameters
        challenge: t.Literal["tls-alpn-01"] = module.params["challenge"]
        challenge_data: dict[str, t.Any] = module.params["challenge_data"]

        # Get hold of private key
        private_key_content_str: str | None = module.params["private_key_content"]
        private_key_passphrase: str | None = module.params["private_key_passphrase"]
        if private_key_content_str is None:
            private_key_content = read_file(module.params["private_key_src"])
        else:
            private_key_content = to_bytes(private_key_content_str)
        try:
            private_key = (
                cryptography.hazmat.primitives.serialization.load_pem_private_key(
                    private_key_content,
                    password=(
                        to_bytes(private_key_passphrase)
                        if private_key_passphrase is not None
                        else None
                    ),
                )
            )
        except Exception as e:
            raise ModuleFailException(f"Error while loading private key: {e}") from e
        if not isinstance(
            private_key,
            (
                cryptography.hazmat.primitives.asymmetric.ed25519.Ed25519PrivateKey,
                cryptography.hazmat.primitives.asymmetric.ed448.Ed448PrivateKey,
                cryptography.hazmat.primitives.asymmetric.rsa.RSAPrivateKey,
                cryptography.hazmat.primitives.asymmetric.dsa.DSAPrivateKey,
                cryptography.hazmat.primitives.asymmetric.ec.EllipticCurvePrivateKey,
            ),
        ):
            raise ModuleFailException(
                f"Cannot use private key type {type(private_key)}"
            )

        # Some common attributes
        domain = to_text(challenge_data["resource"])
        identifier_type, identifier = to_text(
            challenge_data.get("resource_original", "dns:" + challenge_data["resource"])
        ).split(":", 1)
        subject = issuer = cryptography.x509.Name([])
        now = get_now_datetime(with_timezone=CRYPTOGRAPHY_TIMEZONE)
        not_valid_before = now
        not_valid_after = now + datetime.timedelta(days=10)
        san: cryptography.x509.GeneralName
        if identifier_type == "dns":
            san = cryptography.x509.DNSName(identifier)
        elif identifier_type == "ip":
            san = cryptography.x509.IPAddress(ipaddress.ip_address(identifier))
        else:
            raise ModuleFailException(
                f'Unsupported identifier type "{identifier_type}"'
            )

        # Generate regular self-signed certificate
        cert_builder = (
            cryptography.x509.CertificateBuilder()
            .subject_name(subject)
            .issuer_name(issuer)
            .public_key(private_key.public_key())
            .serial_number(cryptography.x509.random_serial_number())
            .add_extension(
                cryptography.x509.SubjectAlternativeName([san]),
                critical=False,
            )
        )
        cert_builder = set_not_valid_before(cert_builder, not_valid_before)
        cert_builder = set_not_valid_after(cert_builder, not_valid_after)
        regular_certificate = cert_builder.sign(
            private_key,
            cryptography.hazmat.primitives.hashes.SHA256(),
        )

        # Process challenge
        if challenge == "tls-alpn-01":
            value = base64.b64decode(challenge_data["resource_value"])
            cert_builder = (
                cryptography.x509.CertificateBuilder()
                .subject_name(subject)
                .issuer_name(issuer)
                .public_key(private_key.public_key())
                .serial_number(cryptography.x509.random_serial_number())
                .add_extension(
                    cryptography.x509.SubjectAlternativeName([san]),
                    critical=False,
                )
                .add_extension(
                    cryptography.x509.UnrecognizedExtension(
                        cryptography.x509.ObjectIdentifier("1.3.6.1.5.5.7.1.31"),
                        encode_octet_string(value),
                    ),
                    critical=True,
                )
            )
            cert_builder = set_not_valid_before(cert_builder, not_valid_before)
            cert_builder = set_not_valid_after(cert_builder, not_valid_after)
            challenge_certificate = cert_builder.sign(
                private_key,
                cryptography.hazmat.primitives.hashes.SHA256(),
            )
        else:
            raise AssertionError("Can never be reached")  # pragma: no cover

        module.exit_json(
            changed=True,
            domain=domain,
            identifier_type=identifier_type,
            identifier=identifier,
            challenge_certificate=challenge_certificate.public_bytes(
                cryptography.hazmat.primitives.serialization.Encoding.PEM
            ),
            regular_certificate=regular_certificate.public_bytes(
                cryptography.hazmat.primitives.serialization.Encoding.PEM
            ),
        )
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

# Copyright: (c) 2022, Ansible Project
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

AMAZON_AWS_COLLECTION_NAME = "amazon.aws"
AMAZON_AWS_COLLECTION_VERSION = "11.3.0"


_collection_info_context = {
    "name": AMAZON_AWS_COLLECTION_NAME,
    "version": AMAZON_AWS_COLLECTION_VERSION,
}


def set_collection_info(collection_name=None, collection_version=None):
    if collection_name:
        _collection_info_context["name"] = collection_name
    if collection_version:
        _collection_info_context["version"] = collection_version


def get_collection_info():
    return _collection_info_context



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Shreyas Srish (@shrsr) <ssrish@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: ndo_ptp_policy_profiles
short_description: Manage PTP Policy Profiles on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage PTP Policy Profiles on Cisco Nexus Dashboard Orchestrator (NDO).
- This module is only supported on ND v3.1 (NDO v4.3) and later.
author:
- Shreyas Srish (@shrsr)
options:
  template:
    description:
    - The name of the template.
    - The template must be a fabric policy template.
    type: str
    required: true
  ptp_policy_profile_name:
    description:
    - The name of the PTP Policy profile.
    type: str
    aliases: [ name ]
  ptp_policy_profile_uuid:
    description:
    - The UUID of the PTP Policy profile.
    type: str
    aliases: [ uuid ]
  delay_interval:
    description:
    - Delay interval of the PTP Policy profile.
    - The value should be between -4 to 4.
    - This is required when O(state=present).
    type: int
  profile_template:
    description:
    - Profile template of the PTP Policy profile.
    - The default value is aes67_2015.
    type: str
    choices: [ aes67_2015, default, telecom_8275_1, smpte_2059_2 ]
  sync_interval:
    description:
    - Sync interval of the PTP Policy profile.
    - The value should be between -4 to -1.
    - This is required when O(state=present).
    type: int
  override_node_profile:
    description:
    - Node profile override of the PTP Policy profile.
    type: bool
  announce_timeout:
    description:
    - Announce timeout of the PTP Policy profile.
    - The value should be between 2 to 10.
    - This is required when O(state=present).
    type: int
  announce_interval:
    description:
    - Announce interval of the PTP Policy profile.
    - The value should be between -3 to 1.
    - This is required when O(state=present).
    type: int
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(template) and ptp_policy object must exist before using this module in your playbook.
  The M(cisco.mso.ndo_template) and M(cisco.mso.ndo_ptp_policy) modules can be used for this.
seealso:
- module: cisco.mso.ndo_template
- module: cisco.mso.ndo_ptp_policy
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Create a new PTP profile
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    ptp_policy_profile_name: ansible_profile
    delay_interval: 2
    sync_interval: 1
    announce_timeout: 3
    announce_interval: 4
    profile_template: aes67_2015
    state: present
  register: create_ptp_profile

- name: Query a PTP profile
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    ptp_policy_profile_name: ansible_profile
    state: query
  register: query_one

- name: Query a PTP profile using UUID
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    ptp_policy_profile_uuid: "{{ create_ptp_profile.current.uuid }}"
    state: query
  register: query_one_uuid

- name: Query all PTP profiles
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    state: query
  register: query_all

- name: Update a PTP profile name with UUID
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    ptp_policy_profile_name: profile_update
    ptp_policy_profile_uuid: "{{ create_ptp_profile.current.uuid }}"

- name: Delete a PTP profile using name
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    ptp_policy_profile_name: ansible_profile
    state: absent

- name: Delete a PTP profile using UUID
  cisco.mso.ndo_ptp_policy_profiles:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    template: ansible_fabric_policy_template
    ptp_policy_profile_uuid: "{{ create_ptp_profile.current.uuid }}"
    state: absent
"""

RETURN = r"""
"""

from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.template import MSOTemplate, KVPair
from ansible_collections.cisco.mso.plugins.module_utils.utils import append_update_ops_data
from ansible_collections.cisco.mso.plugins.module_utils.constants import PROFILE_TEMPLATE
import copy


def main():
    argument_spec = mso_argument_spec()
    argument_spec.update(
        dict(
            template=dict(type="str", required=True),
            ptp_policy_profile_name=dict(type="str", aliases=["name"]),
            ptp_policy_profile_uuid=dict(type="str", aliases=["uuid"]),
            delay_interval=dict(type="int"),
            profile_template=dict(type="str", choices=list(PROFILE_TEMPLATE)),
            sync_interval=dict(type="int"),
            override_node_profile=dict(type="bool"),
            announce_timeout=dict(type="int"),
            announce_interval=dict(type="int"),
            state=dict(type="str", choices=["absent", "query", "present"], default="query"),
        )
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "present", ["ptp_policy_profile_name", "ptp_policy_profile_uuid"], True],
            ["state", "absent", ["ptp_policy_profile_name", "ptp_policy_profile_uuid"], True],
            ["state", "present", ["announce_interval", "announce_timeout", "sync_interval", "delay_interval"]],
        ],
    )

    mso = MSOModule(module)

    template = module.params.get("template")
    ptp_policy_profile_name = module.params.get("ptp_policy_profile_name")
    ptp_policy_profile_uuid = module.params.get("ptp_policy_profile_uuid")
    delay_interval = module.params.get("delay_interval")
    profile_template = PROFILE_TEMPLATE.get(module.params.get("profile_template"))
    sync_interval = module.params.get("sync_interval")
    override_node_profile = module.params.get("override_node_profile")
    announce_timeout = module.params.get("announce_timeout")
    announce_interval = module.params.get("announce_interval")
    state = module.params.get("state")

    ops = []
    match = None

    mso_template = MSOTemplate(mso, "fabric_policy", template)
    mso_template.validate_template("fabricPolicy")
    object_description = "Profiles"

    path = "/fabricPolicyTemplate/template/ptpPolicy/profiles"

    existing_profiles = mso_template.template.get("fabricPolicyTemplate", {}).get("template", {}).get("ptpPolicy", {}).get("profiles") or []

    if ptp_policy_profile_name or ptp_policy_profile_uuid:
        match = mso_template.get_object_by_key_value_pairs(
            object_description,
            existing_profiles,
            [KVPair("uuid", ptp_policy_profile_uuid) if ptp_policy_profile_uuid else KVPair("name", ptp_policy_profile_name)],
        )
        if match:
            mso.existing = mso.previous = copy.deepcopy(match.details)
    else:
        mso.existing = mso.previous = existing_profiles

    if state == "present":
        mso_values = dict(
            name=ptp_policy_profile_name,
            delayIntvl=delay_interval,
            profileTemplate=profile_template,
            syncIntvl=sync_interval,
            nodeProfileOverride=override_node_profile,
            announceTimeout=announce_timeout,
            announceIntvl=announce_interval,
        )

        if match:
            update_path = "{0}/{1}".format(path, match.index)
            append_update_ops_data(ops, match.details, update_path, mso_values)
            mso.sanitize(match.details, collate=True)
        else:
            mso.sanitize(mso_values)
            ops.append(dict(op="add", path="{0}/-".format(path), value=mso.sent))

    elif state == "absent":
        if match:
            ops.append(dict(op="remove", path="{0}/{1}".format(path, match.index)))

    if not module.check_mode and ops:
        response = mso.request(mso_template.template_path, method="PATCH", data=ops)
        profiles = response.get("fabricPolicyTemplate", {}).get("template", {}).get("ptpPolicy", {}).get("profiles") or []
        match = mso_template.get_object_by_key_value_pairs(
            object_description,
            profiles,
            [KVPair("uuid", ptp_policy_profile_uuid) if ptp_policy_profile_uuid else KVPair("name", ptp_policy_profile_name)],
        )
        if match:
            mso.existing = match.details
        else:
            mso.existing = {}
    elif module.check_mode and state != "query":
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright (c) 2021, Cisco Systems
# GNU General Public License v3.0+ (see LICENSE or
# https://www.gnu.org/licenses/gpl-3.0.txt)

DOCUMENTATION = r"""
module: organizations_inventory_onboarding_cloud_monitoring_prepare
short_description: Resource module for organizations _inventory _onboarding _cloud
  _monitoring _prepare
description:
  - Manage operation create of the resource organizations _inventory _onboarding _cloud
    _monitoring _prepare. - > Initiates or updates an import session. An import ID
    will be generated and used when you are ready to commit the import.
version_added: '1.0.0'
extends_documentation_fragment:
  - cisco.meraki.module
author: Francisco Munoz (@fmunoz)
options:
  devices:
    description: A set of devices to import (or update).
    elements: dict
    suboptions:
      sudi:
        description: Device SUDI certificate.
        type: str
      tunnel:
        description: TLS Related Parameters.
        suboptions:
          certificateName:
            description: Name of the configured TLS certificate.
            type: str
          localInterface:
            description: Number of the vlan expected to be used to connect to the
              cloud.
            type: int
          loopbackNumber:
            description: Number of the configured Loopback Interface used for TLS
              overlay.
            type: int
          name:
            description: Name of the configured TLS tunnel.
            type: str
        type: dict
      user:
        description: User parameters.
        suboptions:
          username:
            description: The name of the device user for Meraki monitoring.
            type: str
        type: dict
      vty:
        description: VTY Related Parameters.
        suboptions:
          accessList:
            description: AccessList details.
            suboptions:
              vtyIn:
                description: VTY in ACL.
                suboptions:
                  name:
                    description: Name.
                    type: str
                type: dict
              vtyOut:
                description: VTY out ACL.
                suboptions:
                  name:
                    description: Name.
                    type: str
                type: dict
            type: dict
          authentication:
            description: VTY AAA authentication.
            suboptions:
              group:
                description: Group Details.
                suboptions:
                  name:
                    description: Group Name.
                    type: str
                type: dict
            type: dict
          authorization:
            description: VTY AAA authorization.
            suboptions:
              group:
                description: Group Details.
                suboptions:
                  name:
                    description: Group Name.
                    type: str
                type: dict
            type: dict
          endLineNumber:
            description: Ending line VTY number.
            type: int
          rotaryNumber:
            description: SSH rotary number.
            type: int
          startLineNumber:
            description: Starting line VTY number.
            type: int
        type: dict
    type: list
  options:
    description: Additional options for the import.
    suboptions:
      skipCommit:
        description: Flag to skip adding the device to RDM.
        type: bool
    type: dict
  organizationId:
    description: OrganizationId path parameter. Organization ID.
    type: str
requirements:
  - meraki >= 2.4.9
  - python >= 3.5
seealso:
  - name: Cisco Meraki documentation for organizations createOrganizationInventoryOnboardingCloudMonitoringPrepare
    description: Complete reference of the createOrganizationInventoryOnboardingCloudMonitoringPrepare
      API.
    link: https://developer.cisco.com/meraki/api-v1/#!create-organization-inventory-onboarding-cloud-monitoring-prepare
notes:
  - SDK Method used are
    organizations.Organizations.create_organization_inventory_onboarding_cloud_monitoring_prepare,
  - Paths used are
    post /organizations/{organizationId}/inventory/onboarding/cloudMonitoring/prepare,
"""

EXAMPLES = r"""
- name: Create
  cisco.meraki.organizations_inventory_onboarding_cloud_monitoring_prepare:
    meraki_api_key: "{{ meraki_api_key }}"
    meraki_base_url: "{{ meraki_base_url }}"
    meraki_single_request_timeout: "{{ meraki_single_request_timeout }}"
    meraki_certificate_path: "{{ meraki_certificate_path }}"
    meraki_requests_proxy: "{{ meraki_requests_proxy }}"
    meraki_wait_on_rate_limit: "{{ meraki_wait_on_rate_limit }}"
    meraki_nginx_429_retry_wait_time: "{{ meraki_nginx_429_retry_wait_time }}"
    meraki_action_batch_retry_wait_time: "{{ meraki_action_batch_retry_wait_time }}"
    meraki_retry_4xx_error: "{{ meraki_retry_4xx_error }}"
    meraki_retry_4xx_error_wait_time: "{{ meraki_retry_4xx_error_wait_time }}"
    meraki_maximum_retries: "{{ meraki_maximum_retries }}"
    meraki_output_log: "{{ meraki_output_log }}"
    meraki_log_file_prefix: "{{ meraki_log_file_prefix }}"
    meraki_log_path: "{{ meraki_log_path }}"
    meraki_print_console: "{{ meraki_print_console }}"
    meraki_suppress_logging: "{{ meraki_suppress_logging }}"
    meraki_simulate: "{{ meraki_simulate }}"
    meraki_be_geo_id: "{{ meraki_be_geo_id }}"
    meraki_caller: "{{ meraki_caller }}"
    meraki_use_iterator_for_get_pages: "{{ meraki_use_iterator_for_get_pages }}"
    meraki_inherit_logging_config: "{{ meraki_inherit_logging_config }}"
    devices:
      - sudi: "-----BEGIN CERTIFICATE-----\n        MIIDyTCCArGgAwIBAgIKBBNXOVCGU1YztjANBgkqhkiG9w0BAQsFADAnMQ4wDAYD\n
          \       VQQKEwVDaXNjbzEVMBMGA1UEAxMMQUNUMiBTVURJIENBMB4XDTIxMDUzMTEzNTUx\n
          \       NVoXDTI5MDUxNDIwMjU0MVowbTEpMCcGA1UEBRMgUElEOkM5MjAwTC0yNFAtNEcg\n
          \       U046SkFFMjUyMjBSMksxDjAMBgNVBAoTBUNpc2NvMRgwFgYDVQQLEw9BQ1QtMiBM\n
          \       aXRlIFNVREkxFjAUBgNVBAMTDUM5MjAwTC0yNFAtNEcwggEiMA0GCSqGSIb3DQEB\n
          \       AQUAA4IBDwAwggEKAoIBAQDaUPxW76g...         TR1TuP36bHh13X3vtGiDsCD88Ci2TZIqd/EDkkc7v9ipUUYVVH+YDrPt2Aukb1PH\n
          \       D6K0R+KhgEzRo5x54TlU6oWvjUpwNZUwwdhMWIQaUVkMyZBYNy0jGPLO8jwZhyBg\n
          \       1Fneybr9pwedGbLrAaz+gdEikB8B4a/fvPjVfL5Ngb4QRjFqWuE+X3nLc0kHedep\n
          \       6nfgpUNXMlStVm5nIXKP6OjmzfCHPYh9L2Ehs1TrSk1ser9Ofx0ZMVL/jBZR2EIj\n
          \       OZ8tH6KlX2/B2pbSPIO6kD5c4UA8Cf1...         VR0PAQH/BAQDAgXgMAwGA1UdEwEB/wQCMAAwHwYDVR0jBBgwFoAUSNjx8cJw1Vu7\n
          \       fHMJk6+4uDAD+H8wTQYDVR0RBEYwRKBCBgkrBgEEAQkVAgOgNRMzQ2hpcElEPVVV\n
          \       VUNNaElGcUVFMklFUUVBQWNBQUFBQUFBQUFBQUFBQUFBQUFBQUFBQUE9MB0GA1Ud\n
          \       DgQWBBRdhMkFD/z5hokaQeLbaRsp4hkvbzANBgkqhkiG9w0BAQsFAAOCAQEAMtuh\n
          \       YpBz4xEZ7YdJsLpw67Q0TTJGnTBRpzA...         OwmH/iZ+tDfYQ3W3ElWTW93871DkuW4WQIfbnoHg/F7bF0DKYVkD3rpZjyz3NhzH\n
          \       d7cjTdJXQ85bTAOXDuxKH3qewrXxxOGXgh3I6NUq0UwMTWh84lND7Jl+ZAQkYNS2\n
          \       iHanTZFQBk3ML0NUb7fKDYGRTZRqwQ/upIO4S6LV1cxH/6V0qbMy3sCSHZoMLrW3\n
          \       0m3M6yKpe5+VZzHZwmWdUf3Ot+zKjhveK5/YNsMIASdvtvymxUizq2Hr1hvR/kPc\n
          \       p1vuyWxipU8JfzOh/A==\n        -----END CERTIFICATE-----\n        "
        tunnel:
          certificateName: DeviceSUDI
          localInterface: 1
          loopbackNumber: 1000
          name: MERAKI
        user:
          username: Meraki
        vty:
          accessList:
            vtyIn:
              name: MERAKI_IN
            vtyOut:
              name: MERAKI_OUT
          authentication:
            group:
              name: ''
          authorization:
            group:
              name: MERAKI
          endLineNumber: 17
          rotaryNumber: 50
          startLineNumber: 16
    options:
      skipCommit: false
    organizationId: string
"""
RETURN = r"""
meraki_response:
  description: A dictionary or list with the response returned by the Cisco Meraki Python SDK
  returned: always
  type: list
  sample: >
    [
      {
        "message": "string",
        "udi": "string",
        "deviceId": "string",
        "status": "string",
        "configParams": {
          "tunnel": {
            "mode": "string",
            "port": "string",
            "host": "string",
            "name": "string",
            "rootCertificate": {
              "content": "string",
              "name": "string"
            }
          },
          "cloudStaticIp": "string",
          "user": {
            "publicKey": "string",
            "username": "string",
            "secret": {
              "hash": "string"
            }
          }
        }
      }
    ]
"""



#!/usr/bin/python

# Copyright (c) 2014, Dimitrios Tydeas Mengidis <tydeas.dr@gmail.com>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: composer
author:
  - "Dimitrios Tydeas Mengidis (@dmtrs)"
  - "René Moser (@resmo)"
short_description: Dependency Manager for PHP
description:
  - Composer is a tool for dependency management in PHP. It allows you to declare the dependent libraries your project needs
    and it installs them in your project for you.
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  command:
    type: str
    description:
      - Composer command like V(install), V(update) and so on.
    default: install
  arguments:
    type: str
    description:
      - Composer arguments like required package, version and so on.
    default: ''
  executable:
    type: path
    description:
      - Path to PHP executable on the remote host, if PHP is not in E(PATH).
    aliases: [php_path]
  working_dir:
    type: path
    description:
      - Directory of your project (see C(--working-dir)). This is required when the command is not run globally.
      - This is ignored if O(global_command=true).
  global_command:
    description:
      - Runs the specified command globally.
    type: bool
    default: false
  prefer_source:
    description:
      - Forces installation from package sources when possible (see C(--prefer-source)).
    default: false
    type: bool
  prefer_dist:
    description:
      - Forces installation from package dist even for dev versions (see C(--prefer-dist)).
    default: false
    type: bool
  no_dev:
    description:
      - Disables installation of require-dev packages (see C(--no-dev)).
    default: true
    type: bool
  no_scripts:
    description:
      - Skips the execution of all scripts defined in composer.json (see C(--no-scripts)).
    default: false
    type: bool
  no_plugins:
    description:
      - Disables all plugins (see C(--no-plugins)).
    default: false
    type: bool
  optimize_autoloader:
    description:
      - Optimize autoloader during autoloader dump (see C(--optimize-autoloader)).
      - Convert PSR-0/4 autoloading to classmap to get a faster autoloader.
      - Recommended especially for production, but can take a bit of time to run.
    default: true
    type: bool
  classmap_authoritative:
    description:
      - Autoload classes from classmap only.
      - Implicitly enable optimize_autoloader.
      - Recommended especially for production, but can take a bit of time to run.
    default: false
    type: bool
  apcu_autoloader:
    description:
      - Uses APCu to cache found/not-found classes.
    default: false
    type: bool
  ignore_platform_reqs:
    description:
      - Ignore C(php), C(hhvm), C(lib-*) and C(ext-*) requirements and force the installation even if the local machine does not fulfill
        these.
    default: false
    type: bool
  force:
    description:
      - When O(command) is V(create-project), the module checks whether a V(composer.json) already
        exists in O(working_dir) and skips the command if it does, making the task idempotent.
      - Set to V(true) to always run the command regardless.
    default: false
    type: bool
    version_added: 12.6.0
  composer_executable:
    type: path
    description:
      - Path to composer executable on the remote host, if composer is not in E(PATH) or a custom composer is needed.
    version_added: 3.2.0
requirements:
  - php
  - composer installed in bin path (recommended C(/usr/local/bin)) or specified in O(composer_executable)
notes:
  - Default options that are always appended in each execution are C(--no-ansi), C(--no-interaction) and C(--no-progress)
    if available.
  - We received reports about issues on macOS if composer was installed by Homebrew. Please use the official install method
    to avoid issues.
"""

EXAMPLES = r"""
- name: Download and installs all libs and dependencies outlined in the /path/to/project/composer.lock
  community.general.composer:
    command: install
    working_dir: /path/to/project

- name: Install a new package
  community.general.composer:
    command: require
    arguments: my/package
    working_dir: /path/to/project

- name: Clone and install a project with all dependencies
  community.general.composer:
    command: create-project
    arguments: package/package /path/to/project ~1.0
    working_dir: /path/to/project
    prefer_dist: true

- name: Install a package globally
  community.general.composer:
    command: require
    global_command: true
    arguments: my/package
"""

import os
import re
import shlex

from ansible.module_utils.basic import AnsibleModule


def parse_out(string):
    return re.sub(r"\s+", " ", string).strip()


def has_changed(string):
    return all(
        no_change not in string
        for no_change in ["Nothing to install or update", "Nothing to install, update or remove"]
    )


def get_available_options(module, command="install"):
    # get all available options from a composer command using composer help to json
    rc, out, err = composer_command(module, ["help", command], arguments=["--no-interaction", "--format=json"])
    if rc != 0:
        output = parse_out(err)
        module.fail_json(msg=output)

    command_help_json = module.from_json(out)
    return command_help_json["definition"]["options"]


def composer_command(module, command, arguments=None, options=None):
    if options is None:
        options = []
    if arguments is None:
        arguments = []

    global_command = module.params["global_command"]

    if global_command:
        global_arg = ["global"]
        working_dir_option = []
    else:
        global_arg = []
        working_dir_option = ["--working-dir", module.params["working_dir"]]

    if module.params["executable"] is None:
        php_path = module.get_bin_path("php", True, ["/usr/local/bin"])
    else:
        php_path = module.params["executable"]

    if module.params["composer_executable"] is None:
        composer_path = module.get_bin_path("composer", True, ["/usr/local/bin"])
    else:
        composer_path = module.params["composer_executable"]

    cmd = [php_path, composer_path] + working_dir_option + global_arg + command + options + arguments
    return module.run_command(cmd)


def get_composer_home(module):
    """Get the composer home directory by running 'composer config --global home'."""
    if module.params["executable"] is None:
        php_path = module.get_bin_path("php", True, ["/usr/local/bin"])
    else:
        php_path = module.params["executable"]

    if module.params["composer_executable"] is None:
        composer_path = module.get_bin_path("composer", True, ["/usr/local/bin"])
    else:
        composer_path = module.params["composer_executable"]

    cmd = [php_path, composer_path, "config", "--global", "--no-interaction", "home"]
    rc, out, err = module.run_command(cmd)
    if rc == 0:
        return out.strip()
    return None


def get_config_files(module):
    """Get list of config files that might be changed by 'composer config'."""
    files = ["composer.json", "auth.json"]
    global_command = module.params["global_command"]
    working_dir = module.params["working_dir"]

    if global_command:
        composer_home = get_composer_home(module)
        if composer_home:
            files = [os.path.join(composer_home, f) for f in files]
    elif working_dir:
        files = [os.path.join(working_dir, f) for f in files]

    return files


def hash_config_files(module, files):
    """Return a dict mapping file path to its sha256 hash (or None if the file does not exist)."""
    return {f: (module.sha256(f) if os.path.isfile(f) else None) for f in files}


def main():
    module = AnsibleModule(
        argument_spec=dict(
            command=dict(default="install", type="str"),
            arguments=dict(default="", type="str"),
            executable=dict(type="path", aliases=["php_path"]),
            working_dir=dict(type="path"),
            global_command=dict(default=False, type="bool"),
            prefer_source=dict(default=False, type="bool"),
            prefer_dist=dict(default=False, type="bool"),
            no_dev=dict(default=True, type="bool"),
            no_scripts=dict(default=False, type="bool"),
            no_plugins=dict(default=False, type="bool"),
            apcu_autoloader=dict(default=False, type="bool"),
            optimize_autoloader=dict(default=True, type="bool"),
            classmap_authoritative=dict(default=False, type="bool"),
            ignore_platform_reqs=dict(default=False, type="bool"),
            force=dict(default=False, type="bool"),
            composer_executable=dict(type="path"),
        ),
        required_if=[("global_command", False, ["working_dir"])],
        supports_check_mode=True,
    )
    module.run_command_environ_update = {"LANGUAGE": "C", "LC_ALL": "C"}

    # Get composer command with fallback to default
    command = module.params["command"]
    if re.search(r"\s", command):
        module.fail_json(msg="Use the 'arguments' param for passing arguments with the 'command'")

    arguments = shlex.split(module.params["arguments"])
    available_options = get_available_options(module=module, command=command)

    options = []

    # Default options
    default_options = [
        "no-ansi",
        "no-interaction",
        "no-progress",
    ]

    for option in default_options:
        if option in available_options:
            option = f"--{option}"
            options.append(option)

    option_params = {
        "prefer_source": "prefer-source",
        "prefer_dist": "prefer-dist",
        "no_dev": "no-dev",
        "no_scripts": "no-scripts",
        "no_plugins": "no-plugins",
        "apcu_autoloader": "acpu-autoloader",
        "optimize_autoloader": "optimize-autoloader",
        "classmap_authoritative": "classmap-authoritative",
        "ignore_platform_reqs": "ignore-platform-reqs",
    }

    for param, option in option_params.items():
        if module.params.get(param) and option in available_options:
            option = f"--{option}"
            options.append(option)

    working_dir = module.params["working_dir"]

    if command == "create-project" and not module.params["force"]:
        if working_dir and os.path.exists(os.path.join(working_dir, "composer.json")):
            module.exit_json(changed=False, msg="composer.json already exists in working_dir, skipping create-project")

    if module.check_mode:
        if "dry-run" in available_options:
            options.append("--dry-run")
        else:
            module.exit_json(skipped=True, msg=f"command '{command}' does not support check mode, skipping")

    # For 'config' command in non-check mode, use sha256 hashing to detect changes
    use_hash_detection = command == "config" and not module.check_mode
    if use_hash_detection:
        config_files = get_config_files(module)
        hashes_before = hash_config_files(module, config_files)

    rc, out, err = composer_command(module, [command], arguments, options)

    if rc != 0:
        output = parse_out(err)
        module.fail_json(msg=output, stdout=err)
    else:
        # Composer version > 1.0.0-alpha9 now use stderr for standard notification messages
        output = parse_out(out + err)
        if use_hash_detection:
            hashes_after = hash_config_files(module, config_files)
            changed = hashes_before != hashes_after
        else:
            changed = has_changed(output)
        module.exit_json(changed=changed, msg=output, stdout=out + err)


if __name__ == "__main__":
    main()



#
# -*- coding: utf-8 -*-
# Copyright 2019 Red Hat
# GNU General Public License v3.0+
# (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)#!/usr/bin/python

"""
The nxos vlans fact class
It is in this file the configuration is collected from the device
for a given resource, parsed, and the facts tree is populated
based on the configuration.
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

import ast
import re

from copy import deepcopy

from ansible_collections.ansible.netcommon.plugins.module_utils.network.common import utils
from ansible_collections.ansible.netcommon.plugins.module_utils.network.common.utils import (
    parse_conf_arg,
)

from ansible_collections.cisco.nxos.plugins.module_utils.network.nxos.argspec.vlans.vlans import (
    VlansArgs,
)


class VlansFacts(object):
    """The nxos vlans fact class"""

    def __init__(self, module, subspec="config", options="options"):
        self._module = module
        self.argument_spec = VlansArgs.argument_spec
        spec = deepcopy(self.argument_spec)
        if subspec:
            if options:
                facts_argument_spec = spec[subspec][options]
            else:
                facts_argument_spec = spec[subspec]
        else:
            facts_argument_spec = spec

        self.generated_spec = utils.generate_dict(facts_argument_spec)

    def get_device_data(self, connection, show_cmd):
        """Wrapper method for `connection.get()`
        This exists solely to allow the unit test framework to mock device connection calls.
        """
        return connection.get(show_cmd)

    def populate_facts(self, connection, ansible_facts, data=None):
        """Populate the facts for vlans
        :param connection: the device connection
        :param data: previously collected conf
        :rtype: dictionary
        :returns: facts
        """
        objs = []
        # **TBD**
        # N7K EOL/legacy image 6.2 does not support show vlan | json output.
        # If support is still required for this image then:
        # - Wrapp the json calls below in a try/except
        # - When excepted, use a helper method to parse the run_cfg_output,
        #   using the run_cfg_output data to generate compatible json data that
        #   can be read by normalize_table_data.
        if not data:
            # Use structured for most of the vlan parameter states.
            # This data is consistent across the supported nxos platforms.
            try:
                # Not all devices support | json-pretty but is a workaround for
                # libssh issue https://github.com/ansible/pylibssh/issues/208
                structured = self.get_device_data(connection, "show vlan | json-pretty")
            except Exception:
                # When json-pretty is not supported, we fall back to | json
                structured = self.get_device_data(connection, "show vlan | json")

            # Raw cli config is needed for mapped_vni, which is not included in structured.
            run_cfg_output = self.get_device_data(connection, "show running-config | section ^vlan")
        else:
            running_config = data.split("\n\n")
            structured, run_cfg_output = running_config[0], running_config[1]

        # Create a single dictionary from all data sources
        data = self.normalize_table_data(structured, run_cfg_output)

        for vlan in data:
            obj = self.render_config(self.generated_spec, vlan)
            if obj:
                objs.append(obj)

        ansible_facts["ansible_network_resources"].pop("vlans", None)
        facts = {}
        if objs:
            facts["vlans"] = []
            params = utils.validate_config(self.argument_spec, {"config": objs})
            for cfg in params["config"]:
                facts["vlans"].append(utils.remove_empties(cfg))
        ansible_facts["ansible_network_resources"].update(facts)
        return ansible_facts

    def render_config(self, spec, vlan):
        """
        Render config as dictionary structure and delete keys
          from spec for null values
        :param spec: The facts tree, generated from the argspec
        :param vlan: structured data vlan settings (dict) and raw cfg from device
        :rtype: dictionary
        :returns: The generated config
        Sample inputs: test/units/modules/network/nxos/fixtures/nxos_vlans/show_vlan
        """
        obj = deepcopy(spec)

        obj["vlan_id"] = vlan["vlan_id"]

        # name: 'VLAN000x' (default name) or custom name
        name = vlan["vlanshowbr-vlanname"]
        if name and re.match("VLAN%04d" % int(vlan["vlan_id"]), name):
            name = None
        obj["name"] = name

        # mode: 'ce-vlan' or 'fabricpath-vlan'
        obj["mode"] = vlan["vlanshowinfo-vlanmode"].replace("-vlan", "")

        # enabled: shutdown, noshutdown
        obj["enabled"] = True if "noshutdown" in vlan["vlanshowbr-shutstate"] else False

        # state: active, suspend
        obj["state"] = vlan["vlanshowbr-vlanstate"]

        # non-structured data
        obj["mapped_vni"] = parse_conf_arg(vlan["run_cfg"], "vn-segment")

        return utils.remove_empties(obj)

    def normalize_table_data(self, structured, run_cfg_output):
        """Normalize structured output and raw running-config output into
        a single dict to simplify render_config usage.
        This is needed because:
        - The NXOS devices report most of the vlan settings within two
          structured data keys: 'vlanbrief' and 'mtuinfo', but the output is
          incomplete and therefore raw running-config data is also needed.
        - running-config by itself is insufficient because of major differences
          in the cli config syntax across platforms.
        - Thus a helper method combines settings from the separate top-level keys,
          and adds a 'run_cfg' key containing raw cli from the device.
        """
        # device output may be string, convert to list
        structured = ast.literal_eval(str(structured))

        vlanbrief = []
        mtuinfo = []
        if "TABLE_vlanbrief" in structured:
            # SAMPLE: {"TABLE_vlanbriefid": {"ROW_vlanbriefid": {
            #   "vlanshowbr-vlanid": "4", "vlanshowbr-vlanid-utf": "4",
            #   "vlanshowbr-vlanname": "VLAN0004", "vlanshowbr-vlanstate": "active",
            #   "vlanshowbr-shutstate": "noshutdown"}},
            vlanbrief = structured["TABLE_vlanbrief"]["ROW_vlanbrief"]

            # SAMPLE: "TABLE_mtuinfoid": {"ROW_mtuinfoid": {
            #   "vlanshowinfo-vlanid": "4", "vlanshowinfo-media-type": "enet",
            #   "vlanshowinfo-vlanmode": "ce-vlan"}}
            mtuinfo = structured["TABLE_mtuinfo"]["ROW_mtuinfo"]

        if not isinstance(vlanbrief, list):
            # vlanbrief is not a list when only one vlan is found.
            vlanbrief = [vlanbrief]
            mtuinfo = [mtuinfo]

        # split out any per-vlan cli config
        run_cfg_list = re.split(r"[\n^]vlan ", run_cfg_output)

        # Create a list of vlan dicts where each dict contains vlanbrief,
        # mtuinfo, and non-structured running-config data for one vlan.
        vlans = []
        for index, v in enumerate(vlanbrief):
            v["vlan_id"] = v.get("vlanshowbr-vlanid-utf")
            vlan = {}
            vlan.update(v)
            vlan.update(mtuinfo[index])

            vlan["run_cfg"] = ""
            for item in run_cfg_list:
                # Sample match lines
                # 202\n  name Production-Segment-100101\n  vn-segment 100101
                # 5\n  state suspend\n  shutdown\n  name test-changeme\n  vn-segment 942
                pattern = r"^{0}\s+\S.*vn-segment".format(v["vlan_id"])
                if re.search(pattern, item, flags=re.DOTALL):
                    vlan["run_cfg"] = item
                    break

            vlans.append(vlan)
        return vlans



#!/usr/bin/python
#
# Copyright (c) 2021 Red Hat | Ansible Sakar Mehra<@sakarmehra100@gmail.com | @sakar97>
# Copyright (c) 2019, Vladimir Porshkevich (@porshkevich) <neosonic@mail.ru>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: docker_plugin
short_description: Manage Docker plugins
version_added: 1.3.0
description:
  - This module allows to install, delete, enable and disable Docker plugins.
  - Performs largely the same function as the C(docker plugin) CLI subcommand.
notes:
  - The C(--grant-all-permissions) CLI flag is true by default in this module.
extends_documentation_fragment:
  - community.docker._docker.api_documentation
  - community.docker._attributes
  - community.docker._attributes.actiongroup_docker

attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  idempotent:
    support: full

options:
  plugin_name:
    description:
      - Name of the plugin to operate on.
    required: true
    type: str

  state:
    description:
      - V(absent) remove the plugin.
      - V(present) install the plugin, if it does not already exist.
      - V(enable) enable the plugin.
      - V(disable) disable the plugin.
    default: present
    choices:
      - absent
      - present
      - enable
      - disable
    type: str

  alias:
    description:
      - Local name for plugin.
    type: str
    version_added: 1.8.0

  plugin_options:
    description:
      - Dictionary of plugin settings.
    type: dict
    default: {}

  force_remove:
    description:
      - Remove even if the plugin is enabled.
    default: false
    type: bool

  enable_timeout:
    description:
      - Timeout in seconds.
    type: int
    default: 0

author:
  - Sakar Mehra (@sakar97)
  - Vladimir Porshkevich (@porshkevich)

requirements:
  - "Docker API >= 1.25"
"""

EXAMPLES = r"""
---
- name: Install a plugin
  community.docker.docker_plugin:
    plugin_name: plugin_one
    state: present

- name: Remove a plugin
  community.docker.docker_plugin:
    plugin_name: plugin_one
    state: absent

- name: Enable the plugin
  community.docker.docker_plugin:
    plugin_name: plugin_one
    state: enable

- name: Disable the plugin
  community.docker.docker_plugin:
    plugin_name: plugin_one
    state: disable

- name: Install a plugin with options
  community.docker.docker_plugin:
    plugin_name: weaveworks/net-plugin:latest_release
    plugin_options:
      IPALLOC_RANGE: "10.32.0.0/12"
      WEAVE_PASSWORD: "PASSWORD"
"""

RETURN = r"""
plugin:
  description:
    - Plugin inspection results for the affected plugin.
  returned: success
  type: dict
  sample: {}
actions:
  description:
    - List of actions performed during task execution.
  returned: when O(state) is not V(absent)
  type: list
"""

import traceback
import typing as t

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.community.docker.plugins.module_utils._api import auth
from ansible_collections.community.docker.plugins.module_utils._api.errors import (
    APIError,
    DockerException,
    NotFound,
)
from ansible_collections.community.docker.plugins.module_utils._common_api import (
    AnsibleDockerClient,
    RequestException,
)
from ansible_collections.community.docker.plugins.module_utils._util import (
    DifferenceTracker,
    DockerBaseClass,
)


class TaskParameters(DockerBaseClass):
    plugin_name: str

    def __init__(self, client: AnsibleDockerClient) -> None:
        super().__init__()
        self.client = client
        self.alias: str | None = None
        self.plugin_options: dict[str, t.Any] = {}
        self.debug: bool = False
        self.force_remove: bool = False
        self.enable_timeout: int = 0
        self.state: t.Literal["present", "absent", "enable", "disable"] = "present"

        for key, value in client.module.params.items():
            setattr(self, key, value)


def prepare_options(options: dict[str, t.Any] | None) -> list[str]:
    return (
        [f"{k}={v if v is not None else ''}" for k, v in options.items()]
        if options
        else []
    )


def parse_options(options_list: list[str] | None) -> dict[str, str]:
    return dict(x.split("=", 1) for x in options_list) if options_list else {}


class DockerPluginManager:
    def __init__(self, client: AnsibleDockerClient) -> None:
        self.client = client

        self.parameters = TaskParameters(client)
        self.preferred_name = self.parameters.alias or self.parameters.plugin_name
        self.check_mode = self.client.check_mode
        self.diff = self.client.module._diff
        self.diff_tracker = DifferenceTracker()
        self.diff_result: dict[str, t.Any] = {}

        self.actions: list[str] = []
        self.changed = False

        self.existing_plugin = self.get_existing_plugin()

        state = self.parameters.state
        if state == "present":
            self.present()
        elif state == "absent":
            self.absent()
        elif state == "enable":
            self.enable()
        elif state == "disable":
            self.disable()

        if self.diff:
            self.diff_result["before"], self.diff_result["after"] = (
                self.diff_tracker.get_before_after()
            )

    def get_existing_plugin(self) -> dict[str, t.Any] | None:
        try:
            return self.client.get_json("/plugins/{0}/json", self.preferred_name)
        except NotFound:
            return None
        except APIError as e:
            self.client.fail(to_text(e))

    def has_different_config(self) -> DifferenceTracker:
        """
        Return the list of differences between the current parameters and the existing plugin.

        :return: list of options that differ
        """
        assert self.existing_plugin is not None
        differences = DifferenceTracker()
        if self.parameters.plugin_options:
            settings = self.existing_plugin.get("Settings")
            if not settings:
                differences.add(
                    "plugin_options",
                    parameter=self.parameters.plugin_options,
                    active=settings,
                )
            else:
                existing_options = parse_options(settings.get("Env"))

                for key, value in self.parameters.plugin_options.items():
                    if (
                        (not existing_options.get(key) and value)
                        or not value
                        or value != existing_options[key]
                    ):
                        differences.add(
                            f"plugin_options.{key}",
                            parameter=value,
                            active=existing_options.get(key),
                        )

        return differences

    def install_plugin(self) -> None:
        if not self.existing_plugin:
            if not self.check_mode:
                try:
                    # Get privileges
                    headers = {}
                    registry, dummy_repo_name = auth.resolve_repository_name(
                        self.parameters.plugin_name
                    )
                    header = auth.get_config_header(self.client, registry)
                    if header:
                        headers["X-Registry-Auth"] = header
                    privileges = self.client.get_json(
                        "/plugins/privileges",
                        params={"remote": self.parameters.plugin_name},
                        headers=headers,
                    )
                    # Pull plugin
                    params = {
                        "remote": self.parameters.plugin_name,
                    }
                    if self.parameters.alias:
                        params["name"] = self.parameters.alias
                    response = self.client._post_json(
                        self.client._url("/plugins/pull"),
                        params=params,
                        headers=headers,
                        data=privileges,
                        stream=True,
                    )
                    self.client._raise_for_status(response)
                    for dummy in self.client._stream_helper(response, decode=True):
                        pass
                    # Inspect and configure plugin
                    self.existing_plugin = self.client.get_json(
                        "/plugins/{0}/json", self.preferred_name
                    )
                    if self.parameters.plugin_options:
                        data = prepare_options(self.parameters.plugin_options)
                        self.client.post_json(
                            "/plugins/{0}/set", self.preferred_name, data=data
                        )
                except APIError as e:
                    self.client.fail(to_text(e))

            self.actions.append(f"Installed plugin {self.preferred_name}")
            self.changed = True

    def remove_plugin(self) -> None:
        force = self.parameters.force_remove
        if self.existing_plugin:
            if not self.check_mode:
                try:
                    self.client.delete_call(
                        "/plugins/{0}", self.preferred_name, params={"force": force}
                    )
                except APIError as e:
                    self.client.fail(to_text(e))

            self.actions.append(f"Removed plugin {self.preferred_name}")
            self.changed = True

    def update_plugin(self) -> None:
        if self.existing_plugin:
            differences = self.has_different_config()
            if not differences.empty:
                if not self.check_mode:
                    try:
                        data = prepare_options(self.parameters.plugin_options)
                        self.client.post_json(
                            "/plugins/{0}/set", self.preferred_name, data=data
                        )
                    except APIError as e:
                        self.client.fail(to_text(e))
                self.actions.append(f"Updated plugin {self.preferred_name} settings")
                self.changed = True
        else:
            self.client.fail("Cannot update the plugin: Plugin does not exist")

    def present(self) -> None:
        differences = DifferenceTracker()
        if self.existing_plugin:
            differences = self.has_different_config()

        self.diff_tracker.add(
            "exists", parameter=True, active=self.existing_plugin is not None
        )

        if self.existing_plugin:
            self.update_plugin()
        else:
            self.install_plugin()

        if self.diff or self.check_mode or self.parameters.debug:
            self.diff_tracker.merge(differences)

    def absent(self) -> None:
        self.remove_plugin()

    def enable(self) -> None:
        timeout = self.parameters.enable_timeout
        if self.existing_plugin:
            if not self.existing_plugin.get("Enabled"):
                if not self.check_mode:
                    try:
                        self.client.post_json(
                            "/plugins/{0}/enable",
                            self.preferred_name,
                            params={"timeout": timeout},
                        )
                    except APIError as e:
                        self.client.fail(to_text(e))
                self.actions.append(f"Enabled plugin {self.preferred_name}")
                self.changed = True
        else:
            self.install_plugin()
            if not self.check_mode:
                try:
                    self.client.post_json(
                        "/plugins/{0}/enable",
                        self.preferred_name,
                        params={"timeout": timeout},
                    )
                except APIError as e:
                    self.client.fail(to_text(e))
            self.actions.append(f"Enabled plugin {self.preferred_name}")
            self.changed = True

    def disable(self) -> None:
        if self.existing_plugin:
            if self.existing_plugin.get("Enabled"):
                if not self.check_mode:
                    try:
                        self.client.post_json(
                            "/plugins/{0}/disable", self.preferred_name
                        )
                    except APIError as e:
                        self.client.fail(to_text(e))
                self.actions.append(f"Disable plugin {self.preferred_name}")
                self.changed = True
        else:
            self.client.fail("Plugin not found: Plugin does not exist.")

    @property
    def result(self) -> dict[str, t.Any]:
        plugin_data = {}
        if self.parameters.state != "absent":
            try:
                plugin_data = self.client.get_json(
                    "/plugins/{0}/json", self.preferred_name
                )
            except NotFound:
                # This can happen in check mode
                pass
        result: dict[str, t.Any] = {
            "actions": self.actions,
            "changed": self.changed,
            "diff": self.diff_result,
            "plugin": plugin_data,
        }
        if (
            self.parameters.state == "present"
            and not self.check_mode
            and not self.parameters.debug
        ):
            result["actions"] = None
        return {k: v for k, v in result.items() if v is not None}


def main() -> None:
    argument_spec = {
        "alias": {"type": "str"},
        "plugin_name": {"type": "str", "required": True},
        "state": {
            "type": "str",
            "default": "present",
            "choices": ["present", "absent", "enable", "disable"],
        },
        "plugin_options": {"type": "dict", "default": {}},
        "debug": {"type": "bool", "default": False},
        "force_remove": {"type": "bool", "default": False},
        "enable_timeout": {"type": "int", "default": 0},
    }
    client = AnsibleDockerClient(
        argument_spec=argument_spec,
        supports_check_mode=True,
    )

    try:
        cm = DockerPluginManager(client)
        client.module.exit_json(**cm.result)
    except DockerException as e:
        client.fail(
            f"An unexpected docker error occurred: {e}",
            exception=traceback.format_exc(),
        )
    except RequestException as e:
        client.fail(
            f"An unexpected requests error occurred when trying to talk to the Docker daemon: {e}",
            exception=traceback.format_exc(),
        )


if __name__ == "__main__":
    main()



# Copyright (c) 2018, www.privaz.io Valletech AB
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

# Note that this doc fragment is **PRIVATE** to the collection. It can have breaking changes at any time.
# Do not use this from other collections or standalone plugins/modules!

from __future__ import annotations


class ModuleDocFragment:
    # OpenNebula common documentation
    DOCUMENTATION = r"""
options:
  api_url:
    description:
      - The ENDPOINT URL of the XMLRPC server.
      - If not specified then the value of the E(ONE_URL) environment variable, if any, is used.
    type: str
    aliases:
      - api_endpoint
  api_username:
    description:
      - The name of the user for XMLRPC authentication.
      - If not specified then the value of the E(ONE_USERNAME) environment variable, if any, is used.
    type: str
  api_password:
    description:
      - The password or token for XMLRPC authentication.
      - If not specified then the value of the E(ONE_PASSWORD) environment variable, if any, is used.
    type: str
    aliases:
      - api_token
  validate_certs:
    description:
      - Whether to validate the TLS/SSL certificates or not.
      - This parameter is ignored if E(PYTHONHTTPSVERIFY) environment variable is used.
    type: bool
    default: true
  wait_timeout:
    description:
      - Time to wait for the desired state to be reached before timeout, in seconds.
    type: int
    default: 300
"""
